"""Module with functions to operate processes"""
import asyncio
import subprocess

from asyncio.subprocess import PIPE as APIPE
from subprocess import CalledProcessError, CompletedProcess

NOP = lambda x: None


async def read_stream_and_display(stream, display):
    """Read from stream line by line until EOF, display, and capture the lines."""
    #import IPython; IPython.embed()
    output = []
    while True:
        group = (await stream.read(79))
        if not group:
            break
        output.append(group)
        display(group) # assume it doesn't block
    return b''.join(output)


async def async_run(args, out_display=NOP, err_display=NOP, check=False, **kwargs):
    """Run process asynchronously"""
    process = await asyncio.create_subprocess_exec(
        *args, stdout=APIPE, stderr=APIPE, **kwargs)

    # read child's stdout/stderr concurrently (capture and display)
    try:
        stdout, stderr = await asyncio.gather(
            read_stream_and_display(process.stdout, out_display),
            read_stream_and_display(process.stderr, err_display))
    except Exception:
        process.kill()
        raise
    finally:
        # wait for the process to exit
        retcode = await process.wait()
    if check and retcode:
        raise CalledProcessError(retcode, args,
                                 output=stdout, stderr=stderr)

    return CompletedProcess(args, retcode, stdout, stderr)


def sync_run(args, **kwargs):
    """Run process synchronously"""
    # pylint: disable=subprocess-run-check
    return subprocess.run(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs
    )


def printout(value):
    """Print to stdout"""
    try:
        print(value.strip().decode("utf-8"))
    except UnicodeDecodeError:
        print("printout DecodeError:", value)


def printerr(value):
    """Print to stderr"""
    try:
        print(value.strip().decode("utf-8"))
    except UnicodeDecodeError:
        print("printout DecodeError:", value)



"""Julynter Kernel analysis"""
from .linter import JulynterKernel

COMM = None

def init():
    """Init julynter"""
    # pylint: disable=undefined-variable, global-statement
    global COMM
    COMM = JulynterKernel(get_ipython())
    COMM.register()
    


"""Julynter module"""
import json
import sys
from pathlib import Path

from ._version import __version__

if sys.version_info < (3, 5):
    from .oldcmd import main
else:
    from .handlers import setup_handlers
    from .cmd import main

    HERE = Path(__file__).parent.resolve()

    with (HERE / "labextension" / "package.json").open() as fid:
        data = json.load(fid)


def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": data["name"]
    }]

def _jupyter_server_extension_paths():
    """Register julynter server extension"""
    return [{
        "module": "julynter"
    }]


def _load_jupyter_server_extension(server_app):
    """Registers the API handler to receive HTTP requests from the frontend extension.

    Parameters
    ----------
    server_app: jupyterlab.labapp.LabApp
        JupyterLab application instance
    """
    setup_handlers(server_app.web_app)
    server_app.log.info("Registered HelloWorld extension at URL path /jftemp")

# For backward compatibility with notebook server - useful for Binder/JupyterHub
load_jupyter_server_extension = _load_jupyter_server_extension

if __name__ == "__main__":
    main()



"""Julynter Main CLI"""

import argparse
import sys
import traceback

from .. import util
from . import julynterlab
from . import experiment
from . import run
from . import env
from . import extractpipenv
from . import validate

def main():
    """Julynter Main CLI"""
    parser = argparse.ArgumentParser(description='Lint Jupyter Notebooks')
    subparsers = parser.add_subparsers()

    julynterlab.create_subparsers(subparsers)
    experiment.create_subparsers(subparsers)
    run.create_subparsers(subparsers)
    env.create_subparsers(subparsers)
    extractpipenv.create_subparsers(subparsers)
    validate.create_subparsers(subparsers)

    args, rest = parser.parse_known_args()
    try:
        if not getattr(args, 'func', None):
            parser.print_help()
        else:
            args.func(args, rest)
        if not util.EXITED:
            util.do_exit(0)
    except:  # pylint: disable=bare-except
        if not util.EXITED:
            traceback.print_exc()
            util.do_exit(1)



"""Venv environment definition"""
import shutil

from subprocess import CompletedProcess
from .base import Environment
from .process import async_run


class VenvEnvironment(Environment):
    """Venv environment"""

    def __init__(
            self, command="python -m venv",
            envname=".julynterwork", remove=True, create=True, **kwargs
    ):
        # pylint: disable=too-many-arguments
        super().__init__(create=create, remove=remove, **kwargs)
        self.command = command
        self.envname = envname

    def args(self, cmd):
        cmd_str = [
            "source {}/bin/activate".format(self.envname),
        ]
        cmd_str.append(
            cmd
        )

        return [
            "/bin/bash", "-c",
            " && ".join(cmd_str)
        ]

    def _create_args(self):
        cmd_str = [
            "{} {}".format(self.command, self.envname)
        ]
        return [
            "/bin/bash", "-c",
            " && ".join(cmd_str)
        ]

    async def __aenter__(self):
        if self.create:
            shutil.rmtree(self.envname, ignore_errors=True)
            result = await async_run(self._create_args())
            if result.returncode != 0:
                return (None, result)
            return (self, result)
        return (self, CompletedProcess([], 0, b'', b''))

    async def __aexit__(self, typ, value, traceback):
        if self.remove:
            shutil.rmtree(self.envname, ignore_errors=True)

    def report(self):
        print("VEnvironment: {}".format(self.envname))



"""Define main julynter checks"""
import ast
import os
import sys
import re
import builtins
import traceback
from collections import defaultdict
from copy import copy
import pkg_resources

from .ast_visitors import JulynterImportVisitor, JulynterNameVisitor, JulynterPathVisitor
from .jcomm import JulynterComm


def _julynter_get_package(module_name):
    if module_name not in sys.modules:
        return (2, 'Module {} not imported'.format(module_name))
    module = sys.modules[module_name]
    if not hasattr(module, '__file__'):
        return (0, 'Builtin module')
    package_matches = re.findall('site-packages/(.*?)/', module.__file__)
    if not package_matches:
        return (1, 'Package not found')
    return (-1, package_matches[0])


def _julynter_check_package_version(module_name, requirements):
    status, msg = _julynter_get_package(module_name)
    if status != -1:
        return (status, msg)
    package = msg
    found = None
    if not os.path.exists(requirements):
        return (3, "requirements.txt doesn't exist")
    with open(requirements, 'r') as fil:
        for line in fil:
            if line.strip().startswith(package):
                found = line
                break
    if found is None:
        return (3, 'Module not found on requirements.txt')
    try:
        distribution = pkg_resources.get_distribution(package)
    except pkg_resources.DistributionNotFound:
        return (1, 'Distribution not found')
    version = distribution.version
    req_version = found.strip().split('==')[-1]
    if version != req_version:
        return (3, 'Version mismatch')
    return (0, 'Ok')


class JulynterKernel(JulynterComm):
    """Implements julynter checks"""

    def __init__(self, *args, **kwargs):
        super(JulynterKernel, self).__init__(*args, **kwargs)
        self.history = self.shell.history_manager
        self.cache = {
            'lineno_order': [],
            'python_trees': {},
            'processed': set(),
            'name_definitions': {},
            'import_definitions': {},
            'name_usages': {},
            'cell_dependencies': defaultdict(dict),
            'missing_dependencies': {},

            'last_requirements': '',
            'requirements_checked': {},
            'missing_requirements': defaultdict(dict),
            'requirements_processed': set(),
            'has_imports': [],
        }

    def receive(self, msg):
        data = msg['content']['data']
        operation = data.get('operation', '<undefined>')
        try:
            if operation == 'query':
                req = data.get('requirements', 'requirements.txt')
                self.send(self.julynter_query(req))
            elif operation == 'addModule':
                req = data.get('requirements', 'requirements.txt')
                module = data.get('module', '<undefined>')
                result = self.add_package_to_requirements(module, req)
                if result[0] <= 0:
                    self.send(self.julynter_query(req))
                else:
                    self.send({
                        'operation': 'error',
                        'command': operation,
                        'message': result[1],
                        'errorid': result[0],
                    })
        except Exception:  # pylint: disable=broad-except
            self.send({
                'operation': 'error',
                'command': operation,
                'message': traceback.format_exc()
            })

    def julynter_query(self, requirements_file='requirements.txt'):
        """Extract info from history"""
        self._parse_cell_trees()
        cell_dependencies, missing_dependencies = self._julynter_dependencies()
        has_imports, missing_requirements = self._julynter_imports(requirements_file)
        result = {
            'operation': 'queryResult',
            'executed_code': self._julynter_history(),
            'cell_dependencies': cell_dependencies,
            'missing_dependencies': missing_dependencies,
            'absolute_paths': self._julynter_absolute_paths(),
            'has_imports': has_imports,
            'missing_requirements': missing_requirements,
        }
        return result

    def _parse_cell_trees(self):
        """Parse python cell trees"""
        python_trees = self.cache['python_trees']
        lineno_order = self.cache['lineno_order']
        lineno_order.clear()
        rang = self.history.get_range(raw=False, output=False)
        for _, lineno, inline in rang:
            lineno_order.append(lineno)
            if lineno in python_trees:
                continue
            if inline.startswith("get_ipython().run_cell_magic(\'time\'"):
                inline = inline[42:-2].encode('utf-8').decode('unicode_escape')
            try:
                python_trees[lineno] = ast.parse(inline)
            except SyntaxError:
                python_trees[lineno] = None

    def _name_was_defined_before(self, usage, lineno):
        """Check if name was defined before lineno"""
        name_definitions = self.cache['name_definitions']
        for _, line, _ in reversed(list(self.history.get_range(stop=lineno + 1))):
            if usage in name_definitions.get(line, []):
                return (True, line)
        is_builtin = (
            usage in {'_', '__', '___', '_sh', 'Out', 'In', 'get_ipython'}
            or usage in self.shell.ns_table
            or usage in builtins.__dict__
        )
        return (is_builtin, None)

    def _julynter_dependencies(self):
        """Find dependencies among name usages and definitions"""
        processed = self.cache['processed']
        cell_dependencies = self.cache['cell_dependencies']
        missing_dependencies = self.cache['missing_dependencies']
        import_definitions = self.cache['import_definitions']
        python_trees = self.cache['python_trees']
        lineno_order = self.cache['lineno_order']

        for lineno in lineno_order:
            tree = python_trees.get(lineno, None)
            if lineno in processed or tree is None:
                continue
            processed.add(lineno)
            name_visitor = JulynterNameVisitor()
            name_visitor.visit(tree)
            self.cache['name_definitions'][lineno] = name_visitor.name_definitions
            self.cache['name_usages'][lineno] = name_visitor.name_usages
            import_definitions[lineno] = name_visitor.import_definition
            missing_dependencies[lineno] = []
            for usage in name_visitor.name_usages:
                found, line = self._name_was_defined_before(usage, lineno)
                create_dependency = (
                    found and line is not None
                    and line != lineno and usage not in import_definitions.get(line, [])
                )
                if create_dependency:
                    cell_dependencies[lineno][usage] = line
                if not found:
                    missing_dependencies[lineno].append(usage)
        return cell_dependencies, missing_dependencies

    def _reset_import_cache(self, requirements_file):
        """Reset import cache"""
        self.cache['last_requirements'] = requirements_file
        self.cache['requirements_checked'].clear()
        self.cache['missing_requirements'].clear()
        self.cache['requirements_processed'].clear()
        self.cache['has_imports'].clear()

    def _check_imports(self, imports, lineno, requirements_file):
        """Check imports of a cell in requirements_file"""
        checked = self.cache['requirements_checked']
        missing_requirements = self.cache['missing_requirements']

        for req in imports:
            if checked.get(req, 5) >= 2:
                status, msg = _julynter_check_package_version(req, requirements_file)
                if status >= 2:
                    missing_requirements[lineno][req] = {
                        'status': status,
                        'msg': msg
                    }
                checked[req] = status

    def _julynter_imports(self, requirements_file):
        """Check imports from all cells"""
        missing_requirements = self.cache['missing_requirements']
        processed = self.cache['requirements_processed']
        has_imports = self.cache['has_imports']
        python_trees = self.cache['python_trees']
        lineno_order = self.cache['lineno_order']
        if requirements_file != self.cache['last_requirements']:
            self._reset_import_cache(requirements_file)

        old_missing = copy(missing_requirements)
        missing_requirements.clear()
        # check if old imports are still missing
        for lineno, requirements in old_missing.items():
            self._check_imports(requirements.keys(), lineno, requirements_file)
        # check new imports
        for lineno in lineno_order:
            tree = python_trees.get(lineno, None)
            if lineno in processed or tree is None:
                continue
            processed.add(lineno)
            import_visitor = JulynterImportVisitor()
            import_visitor.visit(tree)
            if import_visitor.imports:
                has_imports.append(lineno)
            self._check_imports(import_visitor.imports, lineno, requirements_file)
        return has_imports, missing_requirements

    def _julynter_history(self):
        """Return dict of executed code by line number"""
        executed_code = {}
        for _, lineno, inline in self.history.get_range(raw=True, output=False):
            executed_code[lineno] = inline
        return executed_code

    def _julynter_absolute_paths(self):
        """Check the existence of absolute paths"""
        python_trees = self.cache['python_trees']
        lineno_order = self.cache['lineno_order']
        absolute_paths = {}
        for lineno in lineno_order:
            tree = python_trees.get(lineno, None)
            if tree is None:
                continue
            path_visitor = JulynterPathVisitor()
            path_visitor.visit(tree)
            if path_visitor.absolute_paths:
                absolute_paths[lineno] = list(path_visitor.absolute_paths)
        return absolute_paths

    def add_package_to_requirements(self, module_name, requirements):
        """Add module to requirements.txt file"""
        # pylint: disable=no-self-use
        module_name = module_name.strip()
        status, msg = _julynter_get_package(module_name)
        if status != -1:
            return (status, msg)
        package = msg
        try:
            distribution = pkg_resources.get_distribution(package)
            version = '=={}'.format(distribution.version)
        except pkg_resources.DistributionNotFound:
            version = ''
        lines = []
        found = False
        if os.path.exists(requirements):
            with open(requirements, 'r') as fil:
                for line in fil:
                    if line.strip().startswith(package):
                        lines.append('{}{}\n'.format(package, version))
                        found = True
                    else:
                        lines.append(line)
        if not found:
            lines.append('{}{}\n'.format(package, version))
        with open(requirements, 'w') as fil:
            fil.writelines(lines)
        return (status, msg)



"""Visit node ast collecting information"""
import ast
import os

class JulynterImportVisitor(ast.NodeVisitor):
    """Collect import names"""
    # pylint: disable=invalid-name

    def __init__(self):
        self.imports = []

    def visit_Import(self, node):
        """Visit Import Node"""
        for alias in node.names:
            self.imports.append(alias.name.split('.')[0])

    def visit_ImportFrom(self, node):
        """Visit ImportFrom Node"""
        if node.module and not node.level:
            self.imports.append(node.module.split('.')[0])


class JulynterNameVisitor(ast.NodeVisitor):
    """Collect name definitions and usages"""
    # pylint: disable=invalid-name

    def __init__(self):
        self.name_definitions = set()
        self.name_usages = set()
        self.import_definition = set()

    def visit_Import(self, node):
        """Visit Import Node"""
        for alias in node.names:
            self.name_definitions.add(alias.asname or alias.name.split('.')[0])
            self.import_definition.add(alias.asname or alias.name.split('.')[0])

    def visit_ImportFrom(self, node):
        """Visit ImportFrom Node"""
        for alias in node.names:
            self.name_definitions.add(alias.asname or alias.name.split('.')[0])
            self.import_definition.add(alias.asname or alias.name.split('.')[0])

    def visit_Name(self, node):
        """Visit Name Node"""
        if isinstance(node.ctx, (ast.Param, ast.Store)):
            self.name_definitions.add(node.id)
        else:
            self.name_usages.add(node.id)

    def visit_FunctionDef(self, node):
        """Visit FunctionDef Node"""
        self.name_definitions.add(node.name)

    def visit_ClassDef(self, node):
        """Visit ClassDef Node"""
        self.name_definitions.add(node.name)

    def visit_arg(self, node):
        """Visit arg Node"""
        self.name_definitions.add(node.arg)


class JulynterPathVisitor(ast.NodeVisitor):
    """Collect absolute paths"""
    # pylint: disable=invalid-name

    def __init__(self):
        self.absolute_paths = set()

    def visit_Str(self, node):
        """Visit Str Node"""
        string = node.s.rstrip('/')
        if os.path.abspath(string) == string and os.path.exists(string) and node.s != '/':
            self.absolute_paths.add(node.s)



"""Conda environment definition"""
import shutil

from pathlib import Path
from subprocess import CompletedProcess, CalledProcessError
from .base import Environment
from .process import async_run


class CondaEnvironment(Environment):
    """Conda environment"""

    def __init__(
            self, operation="--clone base", conda_path="~/anaconda3/",
            envname="julynterwork", remove=True, create=True, **kwargs
    ):
        # pylint: disable=too-many-arguments
        super().__init__(remove=remove, create=create, **kwargs)
        self.operation = operation
        self.anaconda_path = Path(conda_path).expanduser()
        self.envname = envname

    def args(self, cmd):
        cmd_str = [
            ". {}/etc/profile.d/conda.sh".format(self.anaconda_path),
            "conda activate {}".format(self.envname)
        ]
        cmd_str.append(
            cmd
        )

        return [
            "/bin/bash", "-c",
            " && ".join(cmd_str)
        ]

    def _remove_args(self):
        cmd_str = [
            ". {}/etc/profile.d/conda.sh ".format(self.anaconda_path),
            "conda env remove --name {} -y".format(self.envname)
        ]
        return [
            "/bin/bash", "-c",
            " && ".join(cmd_str)
        ]

    def _create_args(self):
        cmd_str = [
            ". {}/etc/profile.d/conda.sh ".format(self.anaconda_path),
            "conda create -y --name {} {}".format(self.envname, self.operation)
        ]
        return [
            "/bin/bash", "-c",
            " && ".join(cmd_str)
        ]

    async def __aenter__(self):
        if self.create:
            await async_run(self._remove_args())
            shutil.rmtree(str(self.anaconda_path / "envs" / self.envname), ignore_errors=True)
            result = await async_run(self._create_args())
            if result.returncode != 0:
                return (None, result)
            return (self, result)
        return (self, CompletedProcess([], 0, b'', b''))

    async def __aexit__(self, typ, value, traceback):
        if self.remove:
            try:
                result = await async_run(self._remove_args())
                if result.returncode != 0:
                    print("WARNING: failed to remove environment {}".format(self.envname))
                    result.check_returncode()
            except CalledProcessError:
                shutil.rmtree(str(self.anaconda_path / "envs" / self.envname), ignore_errors=True)

    def report(self):
        print("Conda env: {}".format(self.envname))



"""This module specifies how to compare cells with normalizers"""
import re
import difflib
import json
import pprint
from copy import deepcopy
from collections import deque
from ..util import vprint, String, Bytes, encode_r, timeout, TimeDecoratorError


class Comparison(object):
    """Base class for Cell Comparison"""
    # pylint: disable=useless-object-inheritance

    def __init__(self, base, result, calculate_similarity, vindex=6):
        self.children = []
        self.base = base
        self.result = result
        self.old_outputs = None
        self.new_outputs = None
        self.calculate_similarity = calculate_similarity
        self.vindex = vindex

    def load(self, old_outputs, new_outputs):
        """Load original and new outputs"""
        self.old_outputs = old_outputs
        self.new_outputs = new_outputs
        return self.old_outputs, self.new_outputs

    def chain(self, comparison):
        """Create a chain of comparisons"""
        self.children.append(comparison)
        return self

    def independent(self):
        """Indicates that the comparison is independent"""
        # pylint: disable=no-self-use
        return True

    def propagate(self, key, value):
        """Propagate diff to dependent children"""
        self.set(key, value)
        for child in self.children:
            if not child.independent():
                child.propagate(key, value)

    def set(self, key, value):
        """Set key of result"""
        self.result[self.base + key] = value

    def get(self, key, default=None):
        """Get value from result"""
        return self.result.get(self.base + key, default)


class Normalizer(Comparison):
    """Base class for normalizations"""

    def normalize(self, obj, name, n=0):
        "Apply normalization"
        # pylint: disable=no-self-use,unused-argument,invalid-name
        return obj

    def load(self, old_outputs, new_outputs):
        self.result[self.base + u'_old_changes'] = 0
        self.result[self.base + u'_new_changes'] = 0
        old_outputs_s = self.normalize(old_outputs, u'_old_changes')
        new_outputs_s = self.normalize(new_outputs, u'_new_changes')
        return super(Normalizer, self).load(old_outputs_s, new_outputs_s)

    def independent(self):
        return (
            self.get(u'_old_changes') != 0
            or self.get(u'_new_changes') != 0
        )


class StringNormalizer(Normalizer):
    """Normalize a string based on a replace"""

    def replace_str(self, obj):
        """Specify normalizer replacement"""
        # pylint: disable=no-self-use
        return obj

    def replace_str_base(self, obj, name):
        """Check if there is a change after the replacement"""
        old = obj
        obj = self.replace_str(obj)
        if old != obj:
            self.result[self.base + name] += 1
        return obj

    def normalize(self, obj, name, n=0):
        space = "  " * n
        if isinstance(obj, list):
            vprint(self.vindex, "{}{}>list({})".format(space, self.base, len(obj)))
            return [self.normalize(x, name, n + 1) for x in obj]
        if isinstance(obj, dict):
            vprint(self.vindex, "{}{}>dict({})".format(space, self.base, len(obj)))
            result = {}
            for k, value in obj.items():
                key = self.normalize(k, name, n + 1)
                vprint(self.vindex + 1, "{}{}>>{}".format(space, self.base, key))
                result[key] = self.normalize(value, name, n + 1)
            return result
        if isinstance(obj, String):
            vprint(self.vindex, "{}{}>str({})".format(space, self.base, len(obj)))
            self.replace_str_base(obj, name)
        return obj


class NormalizeEncode(Normalizer):
    """Transform bytes, list, and dict to UTF-8 strings"""

    def normalize(self, obj, name, n=0):
        space = "  " * n
        if isinstance(obj, list):
            vprint(self.vindex, "{}{}>list({})".format(space, self.base, len(obj)))
            return [self.normalize(x, name, n + 1) for x in obj]
        if isinstance(obj, dict):
            vprint(self.vindex, "{}{}>dict({})".format(space, self.base, len(obj)))
            result = {}
            for k, value in obj.items():
                key = self.normalize(k, name, n + 1)
                vprint(self.vindex + 1, "{}{}>>{}".format(space, self.base, key))
                result[key] = self.normalize(value, name, n + 1)
            return result
        if isinstance(obj, Bytes):
            vprint(self.vindex, "{}{}>bytes({})".format(space, self.base, len(obj)))
            self.result[self.base + name] += 1
            return obj.decode('utf8')
        return obj


class NormalizeExecutionCount(Comparison):
    """Remove execution_count from cell"""

    def load(self, old_outputs, new_outputs):
        old_outputs_c = deepcopy(old_outputs)
        new_outputs_c = deepcopy(new_outputs)

        old_count = []
        new_count = []

        for old in old_outputs_c:
            if u'execution_count' in old:
                old_count.append(old[u'execution_count'])
                del old["execution_count"]

        for new in new_outputs_c:
            if u'execution_count' in new:
                new_count.append(new[u'execution_count'])
                del new['execution_count']


        self.set(u'_old', len(old_count))
        self.set(u'_new', len(new_count))
        self.set(u'_distinct', sum(
            1 for old, new in zip(old_count, new_count)
            if old != new
        ) + abs(len(old_count) - len(new_count)))

        return super(NormalizeExecutionCount, self).load(old_outputs_c, new_outputs_c)

    def independent(self):
        return self.result[self.base + u'_distinct'] != 0


class NormalizeDecimal(StringNormalizer):
    """Cut decimals in the second place"""

    def replace_str(self, obj):
        if u'.' in obj:
            return re.sub(encode_r(r'(\d\.\d\d)(\d*)'), encode_r(r'\1'), obj)
        return obj


class NormalizeDate(StringNormalizer):
    """Remove dates from cells"""

    def replace_str(self, obj):
        if u',' in obj:
            obj = re.sub(encode_r(r'\w\w\w, \d\d \w\w\w \d\d\d\d'), u'1970-01-01T', obj)
        if u'-' in obj or u'\\' in obj or u'/' in obj:
            obj = re.sub(encode_r(r'\d\d?\d?\d?([/\-])\d\d?\1\d\d?\d?\d?T?'), u'1970-01-01T', obj)
            obj = re.sub(encode_r(r'\d\d([/\-])\d\d'), u'01-01', obj)
        return obj


class NormalizeTime(StringNormalizer):
    """Remove time from cells"""

    def replace_str(self, obj):
        if u':' in obj:
            obj = re.sub(
                encode_r(r'\d\d:\d\d:\d\d(\.\d*)?[Zz\+\-]?(\d\d?:\d\d)?'), u'00:00:00', obj
            )
            obj = re.sub(encode_r(r'\d\d:\d\d'), u'00:00', obj)
        return obj


class NormalizeWhitespace(StringNormalizer):
    """Remove extra whitespaces"""

    def replace_str(self, obj):
        obj = re.sub(encode_r(r'\s+'), u' ', obj)
        obj = re.sub(encode_r(r'\s*([!$%^&*()_|~=`{}\[\]:";\'<>?,\/@#])\s*'), encode_r(r'\1'), obj)
        return obj


class NormalizeExceptionPath(StringNormalizer):
    """Remove paths from exceptions"""

    def replace_str(self, obj):
        if u'.py:' in obj:
            obj = re.sub(encode_r(r'\w:\\.*?\.py:\d*'), u'/python/path.py:0', obj)
            obj = re.sub(encode_r(
                r'/(bin|cdrom|etc|lib|mnt|prov|run|srv|sys|usr|boot|dev|home|lib64|'
                r'media|opt|root|sbin|tmp|var|Applications|Library|System|Users)/.*?\.py:\d*'
            ), u'/python/path.py:0', obj)
        return obj


class NormalizeMemory(StringNormalizer):
    """Remove memory addresses"""

    def replace_str(self, obj):
        if u'0x' in obj:
            return re.sub(encode_r(r'at 0x(\d|[A-F]|[a-f])*'), u'at 0x00000000', obj)
        return obj


class NormalizeDeprecation(Normalizer):
    """Remove deprecations"""

    def has_deprecation(self, text):
        """Check if cell has deprecation warning"""
        # pylint: disable=no-self-use
        if isinstance(text, list):
            text = u''.join(text)
        low = text.lower()
        return (u'deprecat' in low) or (u'future' in low)

    def normalize(self, obj, name, n=0):
        result = []
        for out in obj:
            has_deprecation = (
                out.get(u'output_type', u'') == u'error' and (
                    self.has_deprecation(out.get(u'ename', u''))
                    or self.has_deprecation(out.get(u'evalue', u''))
                )
                or out.get(u'output_type', u'') == u'stream' and (
                    self.has_deprecation(out.get(u'text', u''))
                )
            )
            if has_deprecation:
                self.result[self.base + name] += 1
            else:
                result.append(out)
        return result


class NormalizeStream(Normalizer):
    """Combine cell streams"""

    def normalize(self, obj, name, n=0):
        obj = deepcopy(obj)
        result = []
        for out in obj:
            last_output_is_stream = (
                result
                and result[-1].get(u'output_type', u'') == out.get(u'output_type', u'') == u'stream'
                and result[-1].get(u'name', u'') == out.get(u'name', u'')
            )
            if last_output_is_stream:
                result[-1][u'text'] += out.get(u'text', u'')
                self.result[self.base + name] += 1
            else:
                result.append(out)
        return result


class NormalizeImage(Normalizer):
    """Remove images"""

    def normalize(self, obj, name, n=0):
        obj = deepcopy(obj)
        for out in obj:
            if out.get(u'output_type', u'').lower() in {u'display_data', u'execute_result'}:
                data = out.get(u'data', u'')
                if u'image/jpeg' in data:
                    del data[u'image/jpeg']
                    self.result[self.base + name] += 1
                if u'image/png' in data:
                    del data[u'image/png']
                    self.result[self.base + name] += 1
                if u'image/svg+xml' in data:
                    del data[u'image/svg+xml']
                    self.result[self.base + name] += 1
                if u'<svg' in data.get(u'text/html', u'').lower():
                    del data[u'text/html']
                    self.result[self.base + name] += 1
                text_plain = data.get(u'text/plain', u'')
                if isinstance(text_plain, list):
                    text_plain = u'\n'.join(text_plain)
                if (
                        text_plain.startswith(u'<matplotlib.figure')
                        or text_plain.startswith(u'<Figure size')
                ):
                    data[u'text/plain'] = u'<Figure>'
                if text_plain.startswith(u'<matplotlib.text') or text_plain.startswith(u'Text('):
                    data[u'text/plain'] = u'<Text>'
                metadata = out.get(u'metadata', u'')
                if u'needs_background' in metadata:
                    del metadata[u'needs_background']
                    self.result[self.base + name] += 1
        return obj


class NormalizeDataframe(Normalizer):
    """Remove html dataframes"""

    def normalize(self, obj, name, n=0):
        obj = deepcopy(obj)
        for out in obj:
            if out.get(u'output_type', u'').lower() in {u'display_data', u'execute_result'}:
                data = out.get(u'data')
                has_dataframe = (
                    u'class="dataframe"' in data.get(u'text/html', u'').lower()
                    and u'text/plain' in data
                )
                if has_dataframe:
                    del data[u'text/html']
                    self.result[self.base + name] += 1
        return obj


class NormalizeDictionary(StringNormalizer):
    """Sort dict keys"""

    def process_item(self, stack, name):
        """Process a item"""
        item = u''.join(stack[-1].pop())
        strip = item.strip()
        if strip != item:
            self.result[self.base + name] += 1
        stack[-1].append(strip)

    def close_dict(self, stack, current, name, end=u'}'):
        """Finish processing dict"""
        self.process_item(stack, name)
        dic = stack.pop()
        sort = sorted(dic)
        if dic != sort:
            self.result[self.base + name] += 1
        dic = u'{' + u','.join(sort) + end
        if len(stack) == 1:
            current = stack[0]
        else:
            current = stack[-1][-1]
        current.append(dic)
        return current

    def replace_str_base(self, obj, name):
        if u'{' in obj and u'}' in obj:
            current = []
            stack = [current]
            for letter in obj:
                if letter == u'{':
                    current = []
                    stack.append([current])
                elif letter == u'}' and len(stack) > 1:
                    current = self.close_dict(stack, current, name)
                elif letter == u',' and len(stack) > 1:
                    self.process_item(stack, name)
                    current = []
                    stack[-1].append(current)
                else:
                    current.append(letter)

            while len(stack) > 1:
                current = self.close_dict(stack, current, name, end=u'')
            obj = u''.join(stack[0])
        return obj


NORMALIZATIONS = {
    "original": Comparison,
    "encode": NormalizeEncode,
    "execution_count": NormalizeExecutionCount,
    "stream": NormalizeStream,
    "setdict": NormalizeDictionary,
    "dataframe": NormalizeDataframe,
    "exception_path": NormalizeExceptionPath,
    "deprecated": NormalizeDeprecation,
    "whitespace": NormalizeWhitespace,
    "decimal": NormalizeDecimal,
    "date": NormalizeDate,
    "time": NormalizeTime,
    "memory": NormalizeMemory,
    "image": NormalizeImage,
}

DEFAULT_NORMALIZATION = [
    'original', 'encode', 'execution_count', 'stream', 'setdict', 'dataframe',
    'exception_path', 'deprecated', 'whitespace', 'decimal', 'date', 'time',
    'memory', 'image',
]

DEFAULT_SIMILARITY = ['execution_count', 'image']


def iterate(comparison, use_all=False):
    """Iterate in independent comparisons"""
    stack = [comparison]
    while stack:
        current = stack.pop()
        if use_all or current.independent():
            yield current
        for child in current.children:
            stack.append(child)

def flat(outputs):
    """Flatten data structures"""
    result = deque()
    stack = [outputs]
    while stack:
        current = stack.pop()
        if isinstance(current, dict):
            stack.append("{")
            keys = sorted(list(current.keys()))
            for key in keys:
                stack.append(key)
                stack.append(current[key])
            stack.append("}")
        elif isinstance(current, list):
            stack.append("[")
            stack.extend(current)
            stack.append("]")
        else:
            result.appendleft(String(current))
    return ",".join(result)


@timeout(1 * 60, use_signals=False)
def jaccard(first, second):
    """Compare elements using jaccard"""
    first = re.split(r"([^a-zA-Z0-9])", first)
    second = re.split(r"([^a-zA-Z0-9])", second)
    matcher = difflib.SequenceMatcher(None, first, second)
    matches = sum(triple[-1] for triple in matcher.get_matching_blocks())
    length = len(first) + len(second)
    if length:
        return float(matches) / float(length - matches)
    return 1.0


def cell_diff(
        index, old_cell, new_cell, show_report,
        normalizations, calculate_similarity, vindex
):
    """Compare cells"""
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    result = {}
    comparisons = []
    last = None
    outputs = old_cell.get('outputs', []), new_cell.get('outputs', [])
    for name in normalizations:
        if name not in NORMALIZATIONS:
            continue
        vprint(vindex, "Prepare {}".format(name))
        calc = name in calculate_similarity
        comparison = NORMALIZATIONS[name](name, result, calc, vindex + 1)
        comparisons.append(comparison)
        outputs = comparison.load(*outputs)
        if last:
            last.chain(comparison)
        last = comparison
    comp = comparisons[0]

    all_equals = True
    any_equals = False
    for comparison in iterate(comp):
        old_outputs = comparison.old_outputs
        new_outputs = comparison.new_outputs
        vprint(vindex, "Compare {}".format(comparison.base))

        if len(old_outputs) != len(new_outputs):
            comparison.propagate(u"_reason", u"_len")
            comparison.propagate(u"_equals", False)
        else:
            for old, new in zip(old_outputs, new_outputs):
                if set(old.keys()) != set(new.keys()):
                    comparison.propagate(u"_reason", u"_keys")
                    comparison.propagate(u"_equals", False)
                    break
                stop = False
                for key in old.keys():
                    old_value = old[key]
                    new_value = new[key]
                    if old_value != new_value:
                        comparison.propagate(u"_reason", u"_same")
                        comparison.propagate(u"_equals", False)
                        break
                if stop:
                    break

        if comparison.get(u"_equals", True):
            comparison.propagate(u"_reason", None)
            comparison.propagate(u"_equals", True)
            any_equals = True
        else:
            all_equals = False

    for comparison in iterate(comp, use_all=True):
        if comparison.calculate_similarity:
            vprint(vindex, "Similarity {}".format(comparison.base))

            old_flat = (flat(comparison.old_outputs))
            new_flat = (flat(comparison.new_outputs))
            try:
                comparison.set(u"_similar", jaccard(old_flat, new_flat))
                comparison.set(u"_timeout", False)
            except TimeDecoratorError:
                vprint(vindex + 1, "Timeout")
                comparison.set(u"_similar", 0)
                comparison.set(u"_timeout", True)

    if show_report and not all_equals:
        report = []
        temp = []
        print(">", index)
        outjson = "out.{}.json".format(index)
        print("Creating {}".format(outjson))
        with open(outjson, 'w') as fil:
            json.dump(result, fil, indent=2)
        for comparison in iterate(comp, use_all=True):
            reason = comparison.get("_reason")
            if reason:
                old_name = "old." + comparison.base
                new_name = "new." + comparison.base
                if comparison.independent():
                    different_number_of_changes = (
                        result.get(comparison.base + "_old_changes", 0)
                        != result.get(comparison.base + "_new_changes", 0)
                    )
                    if different_number_of_changes:
                        report.append(comparison.base)
                    if result.get(comparison.base + "_distinct", 0):
                        report.append(comparison.base)
                    if temp:
                        print("* {}".format(", ".join(temp)))
                        temp = []
                    print("{} {}".format(comparison.base, reason))
                else:
                    temp.append(comparison.base)

                oldouttxt = "out.{}.{}.{}.txt".format(index, old_name, reason)
                newouttxt = "out.{}.{}.{}.txt".format(index, new_name, reason)
                print("Creating {}".format(oldouttxt))
                print("Creating {}".format(newouttxt))
                with open(oldouttxt, 'w') as fil:
                    fil.write(pprint.pformat(comparison.old_outputs))
                with open(newouttxt, 'w') as fil:
                    fil.write(pprint.pformat(comparison.new_outputs))
            else:
                if comparison.independent():
                    report.append(comparison.base)
                    if temp:
                        print("* {}".format(", ".join(temp)))
                        temp = []
                    print("{} ok".format(comparison.base))
                else:
                    temp.append(comparison.base)
            similarity = comparison.get("_similar", None)
            if similarity is not None:
                temp.append("({})".format(similarity))
        if temp:
            print("* {}".format(", ".join(temp)))
        print(report)
    return all_equals, any_equals, result



"""Manipulate notebook imports"""
import ast
import nbformat
from IPython.core.interactiveshell import InteractiveShell

class ImportVisitor(ast.NodeVisitor):
    """Visit cell source and collect imports"""
    # pylint: disable=invalid-name
    def __init__(self):
        self.imports = []
        self.import_from = []
        self.non_toplevel = []
        self.non_toplevel_from = []
        self.level = 0

    def _extract_name(self, names):
        # pylint: disable=no-self-use
        return [x.name for x in names]

    def visit_Import(self, node):
        """Find cell import"""
        lis = (self.imports if self.level <= 1 else self.non_toplevel)
        lis.extend(self._extract_name(node.names))

    def visit_ImportFrom(self, node):
        """Find cell import from"""
        lis = (self.import_from if self.level <= 1 else self.non_toplevel_from)
        lis.append((node.module, self._extract_name(node.names), node.level))

    def generic_visit(self, node):
        self.level += 1
        super().generic_visit(node)
        self.level -= 1

    def import_only_code(self, toplevel):
        """Return cell code with only imports"""
        result = []
        for imp in self.imports:
            result.append("import {}".format(imp))
        for impf in self.import_from:
            name = "." * impf[2] + (impf[0] or "")
            result.append("from {} import {}".format(name, ", ".join(impf[1])))

        if not toplevel:
            for imp in self.non_toplevel:
                result.append("import {}".format(imp))
            for impf in self.non_toplevel_from:
                name = "." * impf[2] + (impf[0] or "")
                result.append("from {} import {}".format(name, ", ".join(impf[1])))
        return "\n".join(result)


def create_notebook_with_imports(path, toplevel=True):
    """Create notebook that only has imports and return path"""
    try:
        with open(path) as fil:
            notebook = nbformat.read(fil, as_version=4)
        metadata = notebook["metadata"]
    except Exception as exc: # pylint: disable=broad-except
        print("Failed to load notebook {}".format(exc))
        return (None, False)
    language_info = metadata.get("language_info", {})
    language = language_info.get("name", "unknown")
    if language == "python":
        shell = InteractiveShell.instance()
        for cell in notebook['cells']:
            if cell.get("cell_type") == "code":
                source = cell["source"] or ""
                cell["source"] = ""
                try:
                    source = shell.input_transformer_manager.transform_cell(source)
                    if "\0" in source:
                        source = source.replace("\0", "\n")
                    tree = ast.parse(source)
                    visitor = ImportVisitor()
                    visitor.visit(tree)
                    cell["source"] = visitor.import_only_code(toplevel)
                except (IndentationError, SyntaxError):
                    pass
        try:
            with open(path + ".julimp", "w") as fil:
                nbformat.write(notebook, fil)
            return (path + ".julimp", True)
        except Exception as exc: # pylint: disable=broad-except
            print("Failed to save notebook {}".format(exc))
            return (None, False)
    return (path, False)



"""Environment orchestrator"""
import json
from .. import util
from ..util import vprint
from ..runner.runner import clean_diff_result, clean_result, clean_fail
from .process import printout, printerr
from .base import Environment

class EnvironmentOrchestrator:
    """Prepare and run notebook in environment"""

    def __init__(
            self, environment: Environment,
            dryrun_install=False, dryrun_execute=False,
            install_julynter=True, show_output=False,
            initial_verbose=0,
        ):
        # pylint: disable=too-many-arguments
        self.environment: Environment = environment
        self.dryrun_install = dryrun_install
        self.dryrun_execute = dryrun_execute
        self.should_install_julynter = install_julynter
        self.iverbose = initial_verbose
        self.show_output = show_output

        self.install_map = {
            "setup.py": self.install_setups,
            "requirements.txt": self.install_requirements,
            "Pipfile": self.install_pipfiles,
            "Pipfile.lock": self.install_pipfiles
        }

    def rprint(self, num, text):
        """Print in hierarchy"""
        # pylint: disable=no-self-use
        vprint(num, text)

    async def execute_environment(self, cwd, install, notebooks, args):
        """Execute notebookes in environment"""
        async with self.environment as (environment, result):
            if not environment:
                msg = result.stderr.decode('utf-8')
                self.rprint(self.iverbose, "Failed to prepare environment due '{}'".format(msg))
                return {
                    'fail': 'prepare-env',
                    'reason': '<Failed to prepare environment>',
                    'msg': msg
                }

            install_fail = await self.install(cwd, install)
            if install_fail:
                return install_fail
            if notebooks:
                return await self.execute_notebooks(cwd, notebooks, args)
            return {
                'fail': 'no-notebooks',
                'reason': '<no notebooks were specified>',
                'msg': 'prepared environment'
            }

    def _output_data(self, result):
        """Get stdout and stderr from arun"""
        # pylint: disable=no-self-use
        return (
            b"##<>##\nOutput:\n"
            + result.stdout.strip()
            + b"\n##<>##Error:\n"
            + result.stderr.strip()
        )

    def run(self, param, **kwargs):
        """Run param and show results"""
        args = [param]
        if self.show_output:
            args.append(printout)
            args.append(printerr)
        return self.environment.arun(*args, **kwargs)


    async def install_setups(self, cwd, names):
        """Install setup.py files"""
        for name in names:
            if not name:
                continue
            path = (cwd / name).parents[0]
            self.rprint(self.iverbose + 1, "Installing setup {}".format(path))
            result = await self.run(
                "GIT_TERMINAL_PROMPT=0 pip install  --progress-bar off -e '{}'".format(
                    str(path).replace("'", "'\\''")
                ),
            )
            data = self._output_data(result)
            if result.returncode != 0:
                return (False, data)
        return (True, b"")

    async def install_requirements(self, cwd, names):
        """Install requirements.txt files"""
        for name in names:
            if not name:
                continue
            path = (cwd / name)
            self.rprint(self.iverbose + 1, "Installing requirements {}".format(path))
            result = await self.run(
                "GIT_TERMINAL_PROMPT=0 pip install --progress-bar off -r '{}'".format(
                    str(path).replace("'", "'\\''"),
                ),
            )
            data = self._output_data(result)
            if result.returncode != 0:
                return (False, data)
        return (True, b"")

    async def install_pipfiles(self, cwd, names):
        """Install Pipfile files"""
        for name in names:
            if not name:
                continue
            path = (cwd / name)
            self.rprint(self.iverbose + 1, "Converting to requirements.txt: {}".format(path))
            result = await self.run(
                "julynter extractpipenv -p '{}' -o '_julynter_requirements.txt'".format(
                    str(cwd / name)
                ),
            )
            data = self._output_data(result)
            if result.returncode != 0:
                return (False, data)
            result, data = await self.install_requirements(
                path.parents[0], ["_julynter_requirements.txt"],
            )
            if not result:
                return (False, data)
        return (True, b"")

    async def install_julynter(self, _cwd, _names):
        """Install julynter for running notebook"""
        self.rprint(self.iverbose + 1, "Installing julynter")
        result = await self.run(
            "pip install julynter==0.4.0"
        )
        data = self._output_data(result)
        if result.returncode != 0:
            return (False, data)
        return (True, b"")

    async def install(self, ocwd, files):
        """Install dependencies"""
        # pylint: disable=too-many-locals
        self.rprint(self.iverbose, "{}Installing repository dependencies".format(
            "[DRY RUN] " if self.dryrun_install else ""
        ))
        if self.dryrun_install:
            return None
        install_options = []

        if self.should_install_julynter:
            install_options.append(('julynter', self.install_julynter, []))
        for key, names in files.items():
            if key in self.install_map:
                install_options.append((key, self.install_map[key], names))

        self.extra_install_options(install_options)

        cwd = self.environment.project_wd(ocwd)

        installed = True
        data_ok_list = []
        data_failed_list = []
        data_failed = b""
        for spec, func, names in install_options:
            if self.should_exit():
                return
            success, data = await func(cwd, names)
            installed = installed and success
            spec_bytes = spec.encode("utf-8")
            if success:
                data_ok_list.append(spec_bytes)
            else:
                data_failed += b"\n##<<>>##" + spec_bytes + b":\n" + data
                data_failed_list.append(spec_bytes)

        if not installed:
            cause = b"Ok: " + b", ".join(data_ok_list)
            cause += b"\n##<<>>##Failed: " + b", ".join(data_failed_list)
            cause += data_failed
            return {
                'fail': 'install',
                'reason': "<Install Dependency Error>",
                'msg': cause,
            }
        return None

    def extra_install_options(self, install_options):
        """Specify extra dependencies to install"""

    def should_exit(self):
        """Check if it should exit in the middle"""
        # pylint: disable=no-self-use
        return False

    async def execute_notebooks(self, ocwd, notebooks, args):
        """Execute notebooks"""
        self.rprint(self.iverbose, "{}Running {} notebooks".format(
            "[DRY RUN] " if self.dryrun_execute else "",
            len(notebooks)
        ))
        result = {}
        if self.dryrun_execute:
            for notebook in notebooks:
                vprint(
                    2,
                    "[DRY RUN] Would run notebook {} {}"
                    .format(notebook, args)
                )
            result["fail"] = "dryrun"
            return result

        result["notebooks"] = {}
        cwd = self.environment.project_wd(ocwd)
        if "-v" not in args and util.VERBOSE != -1:
            args += " -v {}".format(util.VERBOSE)
        if "-i" not in args:
            args += " -i {}".format(self.iverbose + 1)
        for notebook in notebooks:
            if self.should_exit():
                return result
            self.rprint(self.iverbose, "Running notebook {}".format(notebook))
            presult = await self.run(
                "julynter run '{}' -w ejson {}".format(
                    str(cwd / notebook).replace("'", "'\\''"),
                    args
                ),
            )
            sep = b"<<<<---julyntersep--->>>>"
            if sep in presult.stdout:
                result["notebooks"][notebook] = json.loads(presult.stdout.split(sep)[-1])
            else:
                nresult = {
                    "execution": clean_result(),
                    "fail": clean_fail(),
                    "diff": clean_diff_result(),
                }
                nresult["execution"]["status"] = "error"
                if presult.returncode != 0:
                    nresult["fail"]["reason"] = "<Failed to run notebook>"
                else:
                    nresult["fail"]["reason"] = "<Failed to parse results>"
                nresult["fail"]["msg"] = self._output_data(presult)
                result["notebooks"][notebook] = nresult
        return result



"""Base environment module"""
import asyncio
import os
from subprocess import CompletedProcess
from .process import async_run

class Environment:
    """Base environment class"""

    def __init__(self, *args, remove=True, create=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_outdisplay = lambda x: None
        self.default_errdisplay = lambda x: None
        self.remove = remove
        self.create = create

    def args(self, cmd):
        """Return environment args"""
        # pylint: disable=no-self-use
        return [cmd]

    def arun(self, cmd, out_display=None, err_display=None, **kwargs):
        """Run command in environment asynchronously"""
        out_display = out_display or self.default_outdisplay
        err_display = err_display or self.default_errdisplay
        return async_run(self.args(cmd), out_display, err_display, check=False, **kwargs)

    def run(self, cmd, out_display=None, err_display=None, **kwargs):
        """Run command in environment asynchronously"""
        if os.name == 'nt':
            loop = asyncio.ProactorEventLoop() # for subprocess' pipes on Windows
            asyncio.set_event_loop(loop)
        else:
            loop = asyncio.get_event_loop()
        result = loop.run_until_complete(self.arun(cmd, out_display, err_display, **kwargs))
        return result

    async def __aenter__(self):
        return (self, CompletedProcess([], 0, b'', b''))

    async def __aexit__(self, typ, value, traceback):
        return

    def project_wd(self, cwd):
        """Get project directory in environment"""
        # pylint: disable=no-self-use
        return cwd

    def report(self):
        """Report environment name"""
        # pylint: disable=no-self-use
        print("Environment created")


class DryRunEnvironment(Environment):
    """Dry run command"""

    def arun(self, cmd, out_display=None, err_display=None, **kwargs):
        async def dryrun():
            return CompletedProcess([cmd], 0, b'', b'')
        return dryrun()



"""julynter extract pipenv command"""
import sys
import subprocess
import os
from .. import util
from ..util import do_exit


def extract(args, _):
    """extract pipenv operation"""
    cwd = util.Path(args.path)
    parent = cwd.parents[0]
    requirements_txt = parent / args.output
    with open(str(requirements_txt), "wb") as outf:
        env = dict(os.environ)
        env["GIT_TERMINAL_PROMPT"] = "0"
        result = subprocess.run(
            ["pipenv", "lock", "-r", "--python", sys.executable],
            stdout=outf, cwd=str(parent), env=env, check=False)
        if result.stderr:
            sys.stderr.write(result.stderr)
    do_exit(result.returncode)


def create_subparsers(subparsers):
    """Create subparsers for run command"""
    parser = subparsers.add_parser(
        'extractpipenv', help="Run notebook and check if it reproduces the same results"
    )
    parser.set_defaults(func=extract, command=parser)
    parser.add_argument("-p", "--path", type=str, default="Pipfile.lock",
                        help="pipenv path")
    parser.add_argument("-o", "--output", type=str, default="_julynter_requirements.txt",
                        help="requirements output")



"""Utility module"""
# pylint: disable=unused-import
from __future__ import print_function
import json
import sys

from datetime import date
from threading import Lock

from requests_futures.sessions import FuturesSession
from timeout_decorator import timeout, timeout_decorator
from timeout_decorator import TimeoutError as TimeDecoratorError

from .config import home_config_path, load_project_config
from ._version import __version__


VERBOSE = -1
LOG_LOCK = Lock()
EXITED = False

if sys.version_info < (3, 0):
    TimeoutException = RuntimeError
    String = unicode  # pylint: disable=undefined-variable
    Bytes = str
    encode_r = lambda x: x.decode('utf-8')
else:
    TimeoutException = TimeoutError
    String = str
    Bytes = bytes
    encode_r = lambda x: x


if sys.version_info <= (3, 4):
    from pathlib2 import Path  # pylint: disable=import-error
else:
    from pathlib import Path


def _target(queue, function, *args, **kwargs):
    """Run a function with arguments and return output via a queue.
    This is a helper function for the Process created in _Timeout. It runs
    the function with positional arguments and keyword arguments and then
    returns the function's output by way of a queue. If an exception gets
    raised, it is returned to _Timeout to be raised by the value property.
    """
    try:
        queue.put((True, function(*args, **kwargs)))
    except:  # pylint: disable=bare-except
        #traceback.print_exc()
        queue.put((False, sys.exc_info()[1]))


timeout_decorator._target = _target  # pylint: disable=protected-access



def to_unicode(text):
    """Convert to unicode"""
    if sys.version_info < (3, 0):
        if isinstance(text, unicode):  # pylint: disable=undefined-variable
            return text
        return str(text).decode("utf-8")
    if isinstance(text, str):
        return text
    return bytes(text).decode("utf-8")


def vprint(level, *args):
    """Print according to verbose level"""
    if VERBOSE > level:
        if level > 0:
            print(">" * level, *args)
        else:
            print(*args)


def create_hook(data, folder):
    """Create hook to send experiment data to server"""
    def log_hook(response, *_args, **_kwargs):
        """Hook that sends experiment data to server"""
        newfolder = folder
        try:
            if response.status_code == 200:
                result = response.json()
                if result.get('ok', False):
                    newfolder = "sent_" + folder
                    if "id" in result:
                        data["mid"] = result["id"]
        except Exception as exc:  # pylint: disable=broad-except
            print("Failed to send to server", exc)
        save_log(data, newfolder)
    return log_hook


def log(data, folder="errors", config=None):
    """Log commands to server"""
    config = config or load_project_config()
    if config["experiment"]["enabled"] and config["experiment"]["sendServer"]:
        try:
            session = FuturesSession()
            session.post(config["experiment"]["server"], json={
                'mtype': folder,
                'mtext': data,
                "mexperiment_id": config["experiment"]["id"],
                'version': __version__
            }, hooks={
                'response': create_hook(data, folder),
            })
        except Exception as exc:  # pylint: disable=broad-except
            print("Failed to send to server", exc)
    else:
        save_log(data, folder)


def save_log(data, folder):
    """Save log to folder"""
    today = date.today()
    path = home_config_path() / folder / "{}{}".format(today.year, today.month)
    path.mkdir(parents=True, exist_ok=True)
    with LOG_LOCK:
        try:
            with open(str(path / "{}.log".format(today.day)), 'a') as fil:
                fil.write('\n')
                json.dump(data, fil)
        except json.JSONDecodeError as exc:
            print("Failed to save", data, exc)


def do_exit(exitcode, command=None):
    """Exit program"""
    # pylint: disable=global-statement
    global EXITED
    EXITED = True
    data = {
        "header": "CLI",
        "operation": command or " ".join(sys.argv[1:2]),
        "param": " ".join(sys.argv[1:]),
        "info": str(exitcode),
    }
    log(data, "experiment")
    sys.exit(exitcode)



"""Julynter config utility"""
# pylint: disable=import-outside-toplevel
import json

DEFAULT_EXPERIMENT_SERVER = "https://julynter.npimentel.net"
CONFIG_DIR = '.julynter'


def home_config_path():
    """Return home path"""
    from .util import Path
    return Path.home() / CONFIG_DIR


def load_home_config():
    """Return home confing"""
    data = load_config(home_config_path())
    add_experiment(data)
    return data


def load_project_config(merge_home=True):
    """Return project config"""
    from .util import Path
    project_config = load_config(Path.cwd() / CONFIG_DIR)
    if merge_home:
        home_config = load_home_config()
        project_config = merge(home_config, project_config)
    return project_config


def save_home_config(data):
    """Save home path"""
    return save_config(home_config_path(), data)


def save_project_config(data):
    """Save project path"""
    from .util import Path
    return save_config(Path.cwd() / CONFIG_DIR, data)


def load_config(base):
    """Load julynter config file"""
    data = {}
    try:
        if base.is_dir() and (base / 'config.json').is_file():
            with open(str(base / 'config.json'), 'r') as fil:
                data = json.load(fil)
        elif base.with_suffix('.rc').is_file():
            with open(str(base.with_suffix('.rc')), 'r') as fil:
                data = json.load(fil)
    except json.JSONDecodeError as exc:
        print("Julynter Config ({}) decode error:".format(base), exc)
    return data


def save_config(base, data):
    """Save julynter config file"""
    try:
        if base.with_suffix('.rc').is_file():
            with open(str(base.with_suffix('.rc')), 'w') as fil:
                json.dump(data, fil)
        else:
            base.mkdir(parents=True, exist_ok=True)
            with open(str(base / 'config.json'), 'w') as fil:
                json.dump(data, fil)

        return True
    except json.JSONDecodeError as exc:
        print("Julynter Config ({}) encode error:".format(base), exc)
    return False


def add_experiment(data):
    """Add experiment config to data"""
    if "experiment" not in data:
        data["experiment"] = {}
    exp = data['experiment']
    exp['id'] = exp.get('id', '<unset>')
    exp['lintingMessage'] = exp.get('lintingMessage', False)
    exp['lintingTypes'] = exp.get('lintingTypes', False)
    exp['activity'] = exp.get('activity', False)
    exp['execution'] = exp.get('execution', False)
    exp['code'] = exp.get('code', False)
    exp['name'] = exp.get('name', False)
    exp['enabled'] = exp.get('enabled', False)
    exp['sendServer'] = exp.get('sendServer', False)
    exp["server"] = exp.get("server", DEFAULT_EXPERIMENT_SERVER)


def merge(old, new):
    """Merge dicts"""
    for key, value in new.items():
        if key in old and isinstance(old[key], dict):
            old[key] = merge(old[key], value)
        else:
            old[key] = value
    return old



"""julynter lab command"""
import re
import os
import sys
import pkg_resources


if sys.version_info > (3, 5):
    from jupyterlab import labapp, labextensions
    from jupyterlab.commands import install_extension, build


RESOURCE_NAME = 'julynter'
PACKAGE_NAME = 'julynter-labextension-{}.tgz'
PACKAGE_RE = re.compile('^' + PACKAGE_NAME.format('(.*)\\'))


def version_tuple(version):
    """Get version tuple"""
    parts = version.split('.')
    return tuple(int(x) if x.isnumeric() else x for x in parts)


def install_cmd(_args, _rest):
    """install command"""
    files = pkg_resources.resource_listdir(RESOURCE_NAME, 'labextension')
    versions = [version_tuple(PACKAGE_RE.sub(r'\1', x)) for x in files]
    latest = max(versions)
    latest_file = PACKAGE_NAME.format('.'.join(map(str, latest)))
    resource_file = pkg_resources.resource_filename(RESOURCE_NAME, 'labextension/' + latest_file)
    print('Installing labextension')
    install_extension(resource_file)
    build()


def lab_cmd(_args, rest):
    """lab command"""
    path = pkg_resources.resource_filename(RESOURCE_NAME, 'julynterlab')
    os.environ['JUPYTERLAB_DIR'] = path
    labapp.LabApp.app_dir = os.environ['JUPYTERLAB_DIR']
    labapp.LabApp.launch_instance(argv=rest)


def labextension_cmd(_args, rest):
    """labextension command"""
    path = pkg_resources.resource_filename(RESOURCE_NAME, 'julynterlab')
    os.environ['JUPYTERLAB_DIR'] = path
    labextensions.BaseExtensionApp.app_dir = os.environ['JUPYTERLAB_DIR']
    labextensions.LabExtensionApp.launch_instance(argv=rest)


def create_subparsers(subparsers):
    """create install, lab, and labextension subcommands"""
    ins_parser = subparsers.add_parser(
        'install',
        help="Use nodejs to install the labextension"
    )
    ins_parser.set_defaults(func=install_cmd, command=ins_parser)

    lab_parser = subparsers.add_parser(
        'lab', add_help=False,
        help="Start Jupyter Lab with Julynter"
    )
    lab_parser.set_defaults(func=lab_cmd, command=lab_parser)

    ext_parser = subparsers.add_parser(
        'labextension', add_help=False,
        help="Run Jupyter labextension from Julynter"
    )
    ext_parser.set_defaults(func=labextension_cmd, command=ext_parser)



"""julynter validate command"""
import hashlib
from pathlib import Path
import nbformat
from .. import util
from ..util import vprint, do_exit

GROUP_ATTR = {
    'type': 'reportType',
    'cell': 'cellId'
}

GROUP_MAP = {
    'invalidtitle': 'Invalid Title',
    'hiddenstate': 'Hidden State',
    'confusenotebook': 'Confuse Notebook',
    'import': 'Import',
    'absolutepath': 'Absolute Path'
}


def filter_groups(lint_list):
    """Remove lints"""
    return [x for x in lint_list if x['cellId'] != 'group']


def order_by_type(element):
    """Return ordering for Lint IReport"""
    return (
        element['reportType'],
        element['reportId'],
        element['cellId'],
    )

def order_by_cell(element):
    """Return ordering for Lint IReport"""
    return (
        element['cellId'],
        element['reportType'],
        element['reportId'],
    )


def display_lint(report, extra="", suggestion=True, reason=True):
    """Display lint"""
    vprint(0, "{}{} - {}".format(extra, report['reportId'], report['text']))
    if reason:
        vprint(0, "{}  Reason: {}".format(extra, report['reason']))
    if suggestion:
        vprint(0, "{}  Suggestion: {}".format(extra, report['suggestion']))


def prepare_visible_and_hidden(args, results):
    """Prepare lists of lints and hidden lints"""
    visible = filter_groups(results.get('visible', []))
    filtered_id = filter_groups(results.get('filteredId', []))
    filtered_individual = filter_groups(results.get('filteredIndividual', []))
    filtered_restart = filter_groups(results.get('filteredRestart', []))
    filtered_type = filter_groups(results.get('filteredType', []))
    lints = []
    hidden = []
    (lints if not args.hide_visible else hidden).extend(visible)
    (lints if args.filtered_lint else hidden).extend(filtered_id)
    (lints if args.filtered_message else hidden).extend(filtered_individual)
    (lints if args.filtered_restart else hidden).extend(filtered_restart)
    (lints if args.filtered_type else hidden).extend(filtered_type)
    return lints, hidden


def validate(args, _):
    """Validate notebook"""
    util.VERBOSE = args.verbose
    try:
        with open(args.path) as fil:
            notebook = nbformat.read(fil, as_version=4)
        metadata = notebook["metadata"]
    except Exception as exc: # pylint: disable=broad-except
        vprint(-1, "Failed to load notebook {}".format(exc))
        do_exit(3)

    if 'julynter-results' not in metadata:
        vprint(-1, 'Not valid! You must lint the notebook inside'
                   ' Jupyter Lab before validating it.')
        do_exit(2)

    results = metadata['julynter-results']
    lints, hidden = prepare_visible_and_hidden(args, results)
    exitcode = 0
    if lints:
        lastgroup = None
        usegroup = GROUP_ATTR[args.group] if args.group != "no" else None
        extra = "  " if usegroup else ""
        for lint in lints:
            if usegroup and lint[usegroup] != lastgroup:
                gname = lint[usegroup]
                if lastgroup is not None:
                    vprint(0, "")
                lastgroup = gname
                vprint(0, "{}".format(GROUP_MAP.get(gname, gname)))
            display_lint(
                lint, extra,
                suggestion=not args.hide_explanation,
                reason=not args.hide_explanation,
            )
        vprint(0, "")
        vprint(-1, 'Not valid! Found {} lints'.format(len(lints)))
        exitcode = 1
    else:
        vprint(-1, 'Valid notebook!')

    if not args.hide_warning:
        current = current_filter_letters(args)

        if hidden:
            vprint(-1, '\nWARNING: Found {} hidden lints'.format(len(hidden)))
            vprint(-1, '  Please run "julynter validate -trim <notebook>" to show all hidden lints')
            vprint(-1, '  Or "julynter validate -w{} <notebook>" '
                       'to hide this warning'.format(current))

        if not check_hash(args.path, notebook, results['hash']):
            vprint(-1, '\nWARNING: Hash Mismatch')
            vprint(-1, '  Please perform the linting again in Jupyter Lab')
            vprint(-1, '  Or run "julynter validate -w{} <notebook>" '
                       'to hide this warning'.format(current))

    do_exit(exitcode)


def check_hash(path, notebook, old_hash):
    """Check if current notebook SHA1 matches calculated hash"""
    hashsource = []
    notebook_name = str(Path(path).name).lower()
    hashsource.append(notebook_name + ":")
    for cell in notebook.get('cells', []):
        if cell.get('cell_type', '') == 'code':
            hashsource.append('[{}]'.format(cell.get('execution_count', '') or ''))
        hashsource.append(cell.get('source', '') + ';;;')
    current_hash = hashlib.sha1(''.join(hashsource).encode('utf-8')).hexdigest()
    return current_hash == old_hash


def current_filter_letters(args):
    """Return current filter letters"""
    current = ""
    if args.filtered_type:
        current += "t"
    if args.filtered_restart:
        current += "r"
    if args.filtered_lint:
        current += "i"
    if args.filtered_message:
        current += "m"
    if args.hide_visible:
        current += "d"
    return current


def create_subparsers(subparsers):
    """Create subparsers for validate command"""
    parser = subparsers.add_parser(
        'validate', help=(
            "validate notebook and show existing lints. "
            "It will NOT lint the notebook. "
            "You must use Jupyter Lab instead"
        )
    )
    parser.set_defaults(func=validate, command=parser)
    parser.add_argument(
        "path", type=str,
        help="notebook path")
    parser.add_argument(
        '-i', '--filtered-lint', action="store_true",
        help="Show filtered lint id categories"
    )
    parser.add_argument(
        '-m', '--filtered-message', action="store_true",
        help="Show filtered individual lint messages"
    )
    parser.add_argument(
        '-r', '--filtered-restart', action="store_true",
        help="Show filtered lints that require restarts"
    )
    parser.add_argument(
        '-t', '--filtered-type', action="store_true",
        help="Show filtered lint types"
    )
    parser.add_argument(
        '-d', '--hide-visible', action="store_true",
        help="Ignore visible lints"
    )
    parser.add_argument(
        '-w', '--hide-warning', action="store_true",
        help="Hide warning about hidden lints"
    )
    parser.add_argument(
        '-v', '--verbose', type=int, default=0,
        help="Verbosity level. 0 shows status; 1 shows lints"
    )
    parser.add_argument(
        '-g', '--group', choices=['no', 'type', 'cell'], default='type',
        help="Group lints. Requires -v 1"
    )
    parser.add_argument(
        '-e', '--hide-explanation', action="store_true",
        help="Hide explanation that appear for lints with -v 1"
    )



"""Handle Server extension requests"""
import json

from jupyter_server.base.handlers import APIHandler
from jupyter_server.utils import url_path_join
import tornado
from .config import merge
from .config import load_home_config, load_project_config
from .config import save_home_config, save_project_config
from .util import log


class ProjectConfig(APIHandler):
    """Get and set project config"""
    # pylint: disable=abstract-method
    @tornado.web.authenticated
    def get(self):
        """Get project config"""
        self.finish(json.dumps(load_project_config()))

    @tornado.web.authenticated
    def post(self):
        """Set project config"""
        config = load_project_config(merge_home=False)
        input_data = self.get_json_body()
        if save_project_config(merge(config, input_data)):
            self.finish(json.dumps({'result': 'ok'}))


class UserConfig(APIHandler):
    """Get and set user config"""
    # pylint: disable=abstract-method
    @tornado.web.authenticated
    def get(self):
        """Get user config"""
        self.finish(json.dumps(load_home_config()))

    @tornado.web.authenticated
    def post(self):
        """Set user config"""
        config = load_home_config()
        input_data = self.get_json_body()
        if save_home_config(merge(config, input_data)):
            self.finish(json.dumps({'result': 'ok'}))


class ExperimentData(APIHandler):
    """Send experiment data"""
    # pylint: disable=abstract-method

    @tornado.web.authenticated
    def post(self):
        """Send experiment data"""
        config = load_project_config()

        input_data = self.get_json_body()
        if config['experiment']['enabled']:
            log(input_data, 'experiment', config)


class ErrorData(APIHandler):
    """Send error data"""
    # pylint: disable=abstract-method

    @tornado.web.authenticated
    def post(self):
        """Send error data"""
        input_data = self.get_json_body()
        log(input_data)


def setup_handlers(web_app):
    """Setup notebook handlers"""
    host_pattern = ".*$"

    base_url = web_app.settings["base_url"]
    handlers = [
        (url_path_join(base_url, "julynter", "config"), ProjectConfig),
        (url_path_join(base_url, "julynter", "userconfig"), UserConfig),
        (url_path_join(base_url, "julynter", "experiment"), ExperimentData),
        (url_path_join(base_url, "julynter", "error"), ErrorData),
    ]
    web_app.add_handlers(host_pattern, handlers)



"""Define a Comm for Julynter"""
from ipykernel.comm import Comm

class JulynterComm(object):
    """Julynter comm hadler"""
    # pylint: disable=useless-object-inheritance

    def __init__(self, shell=None):
        self.shell = shell
        self.name = 'julynter.comm'
        self.comm = None

    def register(self):
        """Register comm"""
        self.comm = Comm(self.name)
        self.comm.on_msg(self.receive)
        self.send({'operation': 'init'})

    def receive(self, msg):
        """Receive lint request"""

    def send(self, data):
        """Receive send results"""
        self.comm.send(data)



"""Define commands for Python 2.7"""
import argparse
import traceback

from . import util
from .cmd import run
from .cmd import extractpipenv

def main():
    """Main function"""
    print("This version is not supported! It has limitted analysis features")
    parser = argparse.ArgumentParser(description='Analyze Jupyter Notebooks')
    subparsers = parser.add_subparsers()
    run.create_subparsers(subparsers)
    extractpipenv.create_subparsers(subparsers)

    args, rest = parser.parse_known_args()
    try:
        if not getattr(args, 'func', None):
            parser.print_help()
        else:
            args.func(args, rest)
        if not util.EXITED:
            util.do_exit(0)
    except:  # pylint: disable=bare-except
        if not util.EXITED:
            traceback.print_exc()
            util.do_exit(1)



"""Docker environment definition"""
from copy import copy

from subprocess import CalledProcessError
from pathlib import Path
from .base import Environment
from .process import async_run


class DockerEnvironment(Environment):
    """Docker environment"""

    def __init__(
            self, name, volumes=None, docker_cmd=None,
            cwd=None, wd="/home/jul/execution",
            create=True, remove=True, **kwargs
    ):
        # pylint: disable=too-many-arguments
        super().__init__(create=create, remove=remove, **kwargs)
        self.name = name
        self.wdir = str(wd)
        cwd = cwd or Path.cwd()
        self.volumes = volumes or {str(cwd.expanduser().resolve()): str(wd)}
        self._container_id = None if create else name
        self.docker_cmd = docker_cmd or []

    def _start_container_cmd(self):
        args = ["docker", "run", "-it", "-d"]
        for source, target in self.volumes.items():
            args.append("-v")
            args.append("{}:{}".format(source, target))
        args.append(self.name)
        args.append("/bin/bash")
        return args

    def _kill_container_cmd(self):
        return ["docker", "container", "kill", self._container_id]

    def _rm_container_cmd(self):
        return ["docker", "container", "rm", "-f", self._container_id]

    def _prune_containers_cmd(self):
        # pylint: disable=no-self-use
        return ["docker", "container", "prune", "-f"]

    def args(self, cmd):
        docker_str = copy(self.docker_cmd)
        docker_str.append(
            cmd
        )

        return [
            "docker", "exec", "-t", self._container_id,
            "/bin/bash", "-c",
            " && ".join(docker_str)
        ]

    def use_existing(self, container_id):
        """Use existing container id"""
        self._container_id = container_id
        return self

    async def __aenter__(self):
        if self.create:
            result = await async_run(self._start_container_cmd())
            if result.returncode != 0:
                return (None, result)
            self._container_id = result.stdout.decode('utf-8').strip()
        return (self, result)

    async def __aexit__(self, typ, value, traceback):
        if not self._container_id:
            return
        if self.remove:
            try:
                result = await async_run(self._kill_container_cmd())
                if result.returncode != 0:
                    print("WARNING: failed to kill container {}".format(self._container_id))
                    result.check_returncode()

                result = await async_run(self._rm_container_cmd())
                if result.returncode != 0:
                    print("WARNING: failed to remove container {}. Trying to prune it"
                          .format(self._container_id))
                    result.check_returncode()

            except CalledProcessError:
                result = await async_run(self._prune_containers_cmd())
                if result.returncode != 0:
                    print("WARNING: failed to remove container {}. Trying to prune it"
                          .format(self._container_id))

    def project_wd(self, cwd):
        """Get project directory in environment"""
        return Path(self.wdir)

    def report(self):
        print("Docker container: {}".format(self._container_id))



"""julynter env command"""
import asyncio
import argparse
import os
import sys
from pathlib import Path
from .. import util
from ..environments.docker import DockerEnvironment
from ..environments.conda import CondaEnvironment
from ..environments.venv import VenvEnvironment
from ..environments.orchestrator import EnvironmentOrchestrator
from ..environments.imports import create_notebook_with_imports
from ..cmd.run import simple_view, add_run_arguments
from ..util import do_exit


def discover_files(cwd, values, pattern, exclude=None):
    """Discover files in directory"""
    if "<discover>" not in values:
        return values
    index = values.index("<discover>")
    values.pop(index)
    for name in cwd.glob(pattern):
        name = str(name.relative_to(cwd))
        if exclude and exclude in name:
            continue
        if name not in values:
            values.insert(index, name)
    return values


def create_environment(args, cwd, create, remove):
    """Create environment based on arguments"""
    if args.env_type == "docker":
        return DockerEnvironment(
            args.image_name, cwd=cwd,
            wd=args.target_dir, create=create, remove=remove
        )
    if args.env_type == "conda":
        return CondaEnvironment(
            operation=args.create_operation, conda_path=args.conda_path,
            envname=args.envname, create=create, remove=remove
        )
    #if args.env_type == "venv":
    return VenvEnvironment(
        command=args.create_operation, envname=args.envname,
        create=create, remove=remove
    )


def prepare_files(cwd, args):
    """Discover and preprocess files for running"""
    install = {}
    toremove = []
    notebooks = []

    install = {}
    install['setup.py'] = discover_files(cwd, args.setups, "**/setup.py")
    install['requirements.txt'] = discover_files(cwd, args.requirements, "**/requirements.txt")
    install['Pipfile'] = discover_files(cwd, args.pipfiles, "**/Pipfile")
    install['Pipfile.lock'] = discover_files(cwd, args.pipfile_locks, "**/Pipfile.lock")
    discnotebooks = discover_files(cwd, args.notebooks, "**/*.ipynb", '.ipynb_checkpoints')
    targs = " ".join(args.execution_options or [])
    if args.mode in ("import", "importtop"):
        for path in discnotebooks:
            npath, newfile = create_notebook_with_imports(
                path, toplevel=args.mode == "importtop"
            )
            if npath:
                notebooks.append(npath)
            if newfile:
                toremove.append(npath)
        if "-x" not in targs:
            targs += " -x" # skip-comparison
    elif args.mode == "run":
        notebooks = discnotebooks
    return install, notebooks, targs, toremove


def display_execution_results(res, args, eargs, exitcode):
    """Display execution results"""
    if args.view_mode == "simple":
        print("Report:")
    fails = 0
    skips = 0
    same = 0
    same_norm = 0
    for name, nresult in res["notebooks"].items():
        if args.view_mode == "simple":
            print("  {}".format(name))
            simple_view(eargs, nresult, spaces=4)
        has_error = (
            not nresult.get("execution", None)
            or nresult['execution']['status'] == 'error'
            or 'timeout' in nresult['execution']['processed']
            or 'exception' in nresult['execution']['processed']
            or (
                not eargs.skip_comparison
                and not 'finished' in nresult.get('diff', {}).get('processed', [])
            )
        )
        if has_error:
            exitcode = 3
            fails += 1
        did_skip = (not has_error and (
            'not-run' in nresult['execution']['processed']
            or 'skipped' in nresult['execution']['processed']
        ))
        if did_skip:
            skips += 1

        if not has_error and not did_skip and not eargs.skip_comparison:
            if nresult["diff"]["diff_count"] == 0:
                same += 1
            if nresult["diff"]["diffnorm_count"] == 0:
                same_norm += 1

    print("\nSummary:")
    print("  Total: {} notebooks. Skips: {}; Fails: {}".format(
        len(res['notebooks']), skips, fails
    ))
    if not eargs.skip_comparison:
        print("  Same results before normalizations: {} notebooks".format(same))
        print("  Same results after normalization: {} notebooks".format(same_norm))

    return exitcode


def display_results(res, args, eargs):
    """Display general results"""
    fail = res.get('fail', False)
    exitcode = 0
    if fail == 'prepare-env':
        exitcode = 1
        print("Failed to prepare environment")
        print("  Reason: {}".format(res['reason']))
        if not args.hide_message:
            print("  Message: {}".format(res['msg']))
    elif fail == 'install':
        exitcode = 2
        print("Failed to install declared dependencies")
        print("  Reason: {}".format(res['reason']))
        if not args.hide_message:
            print("  Message: {}".format(res['msg']))
    elif fail == 'no-notebooks':
        if args.mode == "prepare":
            print("Environment prepared")
        else:
            print("Failed to run notebooks in a prepared environment: no notebooks")
    elif fail == 'dryrun':
        print("Finished dry-run")
    else:
        exitcode = display_execution_results(res, args, eargs, exitcode)
    return exitcode

async def aenv(args, eargs):
    """run operation"""
    cwd = Path(args.dir)

    remove = args.mode != "prepare"
    envi = create_environment(args, cwd, create=True, remove=remove)

    runner = EnvironmentOrchestrator(
        envi, initial_verbose=args.initial_verbose,
        install_julynter=not args.dont_install_julynter,
        show_output=args.full_outputs
    )
    install, notebooks, targs, toremove = prepare_files(cwd, args)
    exitcode = 0
    try:
        res = await runner.execute_environment(cwd, install, notebooks, targs)
        exitcode = display_results(res, args, eargs)
    finally:
        if not remove:
            envi.report()
        for path in toremove:
            os.remove(path)
    do_exit(exitcode)

def env(args, _):
    """run operation"""
    parser = argparse.ArgumentParser(description='Run notebook arguments')
    add_run_arguments(parser)
    eargs = parser.parse_args(args.execution_options or [])
    eargs.hide_message = args.hide_message
    util.VERBOSE = args.verbose
    if os.name == 'nt':
        loop = asyncio.ProactorEventLoop() # for subprocess' pipes on Windows
        asyncio.set_event_loop(loop)
    else:
        loop = asyncio.get_event_loop()
    loop.run_until_complete(aenv(args, eargs))


def create_subparsers(subparsers):
    """Create subparsers for run command"""
    parser = subparsers.add_parser(
        'env', help="Run notebooks in a custom environment"
    )

    envparser = argparse.ArgumentParser(add_help=False)
    envparser.add_argument(
        'notebooks', type=str, nargs="*", default=["<discover>"],
        help="notebook paths",
    )
    envparser.add_argument(
        '-d', '--dir', type=str, default=os.getcwd(),
        help="base project dir",
    )

    envparser.add_argument(
        "-j", "--dont-install-julynter", action="store_true",
        help="do not install julynter in environment"
    )
    envparser.add_argument(
        "-s", "--setups", type=str, nargs="*", default=["<discover>"],
        help="setup.py paths"
    )
    envparser.add_argument(
        "-r", "--requirements", type=str, nargs="*", default=["<discover>"],
        help="requirements.txt paths"
    )
    envparser.add_argument(
        "-p", "--pipfiles", type=str, nargs="*", default=["<discover>"],
        help="Pipfile paths"
    )
    envparser.add_argument(
        "-l", "--pipfile-locks", type=str, nargs="*", default=["<discover>"],
        help="Pipfile.lock paths"
    )
    envparser.add_argument(
        "-i", "--initial-verbose", type=int, default=0,
        help="initial verbose level"
    )
    envparser.add_argument(
        "-v", "--verbose", type=int, default=-1,
        help="increase output verbosity")
    envparser.add_argument(
        "-o", "--execution-options", nargs=argparse.REMAINDER,
        help="options for 'julynter run' command")
    envparser.add_argument(
        "-m", "--hide-message", action="store_true",
        help="hide error messages"
    )
    envparser.add_argument(
        "-f", "--full-outputs", action="store_true",
        help="display output of subprocesses"
    )
    envparser.add_argument(
        "-w", "--view-mode", type=str, choices=["simple", "summary"], default="simple",
        help="result visualization mode"
    )

    parser.add_argument(
        "mode", choices=["run", "prepare", "import", "importtop"],
        help="execution modes"
    )
    subp = parser.add_subparsers()


    docker = subp.add_parser(
        'docker', parents=[envparser],
        help="use docker environment"
    )
    docker.set_defaults(env_type="docker")

    docker.add_argument(
        '-e', '--image-name', type=str, default="python",
        help="image name"
    )
    docker.add_argument(
        '-t', '--target-dir', type=str, default="/home/jul/execution",
        help="target execution dir"
    )

    conda = subp.add_parser(
        'conda', parents=[envparser],
        help='use conda environment'
    )
    conda.set_defaults(env_type="conda")
    conda.add_argument(
        '-e', '--envname', type=str, default="julynterwork",
        help="target environment name"
    )
    conda.add_argument(
        '-a', '--conda-path', type=str, default="~/anaconda3",
        help="conda base path"
    )
    conda.add_argument(
        '-c', '--create-operation', type=str, default="python={}.{}".format(*sys.version_info),
        help="create command arguments. "
        "You may set it as '--clone base' if you want to clone an existing environment"
    )

    venv = subp.add_parser(
        'venv', parents=[envparser],
        help='use venv environment'
    )
    venv.set_defaults(env_type="venv")
    venv.add_argument(
        '-e', '--envname', type=str, default=".julynterwork",
        help="target venv name"
    )
    venv.add_argument(
        '-c', '--create-operation', type=str, default="python -m venv",
        help="create venv command. "
        "You may set it 'virtualenv venv --python=python{}.{}'".format(*sys.version_info)
    )

    parser.set_defaults(func=env, command=envparser)



"""Define execution order"""
# pylint: disable=abstract-method, attribute-defined-outside-init, no-member
from nbconvert.preprocessors import ExecutePreprocessor, Preprocessor

from ..util import vprint


class UnsafePreprocessor(Preprocessor):
    """Run cells following the order specified in cell_order. Define the last_try"""

    def safety_fix(self, notebook, index):
        """Fix cell code to run it safer"""

    def preprocess(self, nb, resources):
        for order, index in enumerate(self.cell_order):
            vprint(self.vindex, "{}- Running cell {}".format(order, index))
            self.safety_fix(nb, index)
            self.last_try = (order, index)
            nb.cells[index], resources = self.preprocess_cell(
                nb.cells[index], resources, index
            )
        return nb, resources


class SafePreprocessor(UnsafePreprocessor):
    """Replace unsafe commands from cell"""

    def safety_fix(self, notebook, index):
        has_kill = (
            notebook.cells[index].get('cell_type', '') == 'code'
            and '!kill -9 -1' in notebook.cells[index].get('source', '')
        )
        if has_kill:
            notebook.cells[index]['source'] = (
                notebook.cells[index]['source']
                .replace('!kill -9 -1', '')
            )


class TopBottomAllCellsPreprocessor(Preprocessor):
    """Run all cells in TopDown order"""

    def prepare_notebook_order(self, notebook, vindex):
        """Define the execution of all cells"""
        self.vindex = vindex
        self.cell_order = list(range(len(notebook.cells)))
        return ""


class TopBottomPreprocessor(Preprocessor):
    """Run cells with execution count in TopDown order"""

    def prepare_notebook_order(self, notebook, vindex):
        """Define the execution of cells with execution count"""
        self.vindex = vindex
        cells = [
            index for index, cell in enumerate(notebook.cells)
            if isinstance(cell.get(u'execution_count'), int)
        ]
        self.cell_order = cells
        if not cells:
            return u"No numbered cells"
        return ""


class ExecutionCountPreprocessor(Preprocessor):
    """Run cells following the execution count"""

    def prepare_notebook_order(self, notebook, vindex):
        """Define the execution of cells following the execution count"""
        self.vindex = vindex
        self.cell_order = []
        cells = sorted([
            (int(cell.get(u'execution_count')), index, cell)
            for index, cell in enumerate(notebook.cells)
            if isinstance(cell.get(u'execution_count'), int)
        ])
        if not cells:
            return u"No numbered cells"

        numbers = {count for count, _, _ in cells}
        if len(numbers) != len(cells):
            return u"Repeated cell numbers"

        self.cell_order = [index for _, index, _ in cells]
        return ""


def create_preprocessor(order, unsafe, vindex):
    """Create preprocessor class based on the execution order and safeness"""
    # pylint: disable=too-many-ancestors
    if order.lower() in ("a", "all", "0"):
        vprint(vindex, "TopDown Order - All cells")
        middle = TopBottomAllCellsPreprocessor
    elif order.lower() in ("e", "ec", "executioncount", "1"):
        vprint(vindex, "ExecutionCount Order")
        middle = ExecutionCountPreprocessor
    else: #order.lower() in ("t", "topdown", "td", "2"):
        vprint(vindex, "TopDown Order")
        middle = TopBottomPreprocessor
    last = UnsafePreprocessor if unsafe else SafePreprocessor
    class NotebookPreprocessor(ExecutePreprocessor, middle, last):
        """Resulting preprocessor"""
    return NotebookPreprocessor



"""julynter run command"""
import json
from .. import util
from ..runner.runner import Runner
from ..runner.compare import NORMALIZATIONS, DEFAULT_NORMALIZATION, DEFAULT_SIMILARITY


def json_view(runner):
    """Display result as json"""
    print(json.dumps(runner, indent=2))


def simple_view(args, runner, spaces=2):
    """Display summarized result as prints"""
    print("{}Status: {}".format(" " * spaces, runner['execution']["status"]))
    print("{}Processed: {}".format(" " * spaces, runner['execution']["processed"]))
    if runner['execution']["status"] == "run":
        select_cells = list(zip(
            range(runner['execution']["executed_cells"]),
            runner['execution']["cell_order"]
        ))
        print("{}  Cells order: {}".format(" " * spaces, ", ".join(
            str(index) for _, index in select_cells
        )))
        print("{}  Duration: {}".format(" " * spaces, runner['execution']["duration"]))
    if runner['fail']['msg']:
        print("{}  Reason: {}".format(" " * spaces, runner['fail']['reason']))
        if not args.hide_message:
            print("{}  Message: {}".format(" " * spaces, runner['fail']['msg']))
    if "finished" in runner["diff"]['processed'] and not args.skip_comparison:
        print("{}  Diff:".format(" " * spaces))
        norms = {x: [] for x in args.normalizations}
        for sim in runner['diff']['similarities']:
            for norm in args.normalizations:
                if not sim.get(norm + '_equals', True):
                    norms[norm].append(str(sim['index']))
        anydiff = False
        for norm in args.normalizations:
            if norms[norm]:
                anydiff = True
                print("{}    {}: {}".format(" " * spaces, norm, ', '.join(norms[norm])))
        if not anydiff:
            print("{}    No diff".format(" " * spaces))


def run(args, _):
    """run operation"""
    util.VERBOSE = args.verbose
    runner = Runner(
        args.path, args.cell_order, args.unsafe,
        args.kernel, args.force_fail, args.timeout,
        args.show_report,
        args.normalizations, args.calculate_similarity,
        args.initial_verbose
    )
    finished_run = runner.run()
    if finished_run and not args.skip_comparison:
        runner.compare()
    if args.output:
        runner.save(args.output)
    dictresult = {
        'fail': runner.fail,
        'execution': runner.result,
        'diff': runner.diff_result
    }
    if args.view_mode in ("ejson", "json"):
        if args.view_mode == "ejson":
            print("<<<<---julyntersep--->>>>")
        json_view(dictresult)
    elif args.view_mode == "simple":
        simple_view(args, dictresult)


def create_subparsers(subparsers):
    """Create subparsers for run command"""
    runparser = subparsers.add_parser(
        'run', help="Run notebook and check if it reproduces the same results"
    )
    runparser.set_defaults(func=run, command=runparser)
    add_run_arguments(runparser)
    runparser.add_argument(
        "path", type=str,
        help="notebook path")

def add_run_arguments(runparser):
    """Add run arguments to parsers"""
    runparser.add_argument(
        "-c", "--cell-order", type=str, default="t", choices=[
            '0', 'a', 'all',
            '1', 'e', 'ec', 'executioncount',
            '2', 't', 'td', 'topdown'
        ],
        help=(
            "cell execution order: "
            "'a' - top down, all cells; "
            "'t' - top down, cells with execution count; "
            "'e' - execution count order"
        )
    )
    runparser.add_argument(
        "-u", "--unsafe", action="store_false",
        help="disable filtering some unsafe code")
    runparser.add_argument(
        "-v", "--verbose", type=int, default=-1,
        help="increase output verbosity")
    runparser.add_argument(
        "-k", "--kernel", type=str,
        help="specify Jupyter kernel. By default it tries to read from the notebook"
    )
    runparser.add_argument(
        "-f", "--force-fail", action="store_true",
        help="do not fallback to the default behavior on invalid parameters"
    )
    runparser.add_argument(
        "-r", "--show-report", action="store_true",
        help="show mismatch report"
    )
    runparser.add_argument(
        "-t", "--timeout", type=float, default=300,
        help="notebook timeout time (in seconds)")
    runparser.add_argument(
        "-n", "--normalizations", nargs="+", choices=NORMALIZATIONS.keys(),
        default=DEFAULT_NORMALIZATION,
        help="normalization order"
    )
    runparser.add_argument(
        "-s", "--calculate-similarity", nargs="+", choices=NORMALIZATIONS.keys(),
        default=DEFAULT_SIMILARITY,
        help="calculate post-normalization similarity"
    )
    runparser.add_argument(
        "-o", "--output", type=str,
        help="output notebook"
    )
    runparser.add_argument(
        "-x", "--skip-comparison", action="store_true",
        help="do not compare results after execution"
    )
    runparser.add_argument(
        "-i", "--initial-verbose", type=int, default=1,
        help="initial verbose level"
    )
    runparser.add_argument(
        "-w", "--view-mode", type=str, choices=["ejson", "json", "simple"], default="simple",
        help="result visualization mode"
    )
    runparser.add_argument(
        "-m", "--hide-message", action="store_true",
        help="hide error messages"
    )



"""Manage experiment"""
import zipfile
import os
from datetime import datetime
from ..config import load_home_config, save_home_config, home_config_path
from ..config import DEFAULT_EXPERIMENT_SERVER
from ..util import log


def base_experiment_cmd(args, rest):
    """Execute experiment function"""
    if not getattr(args, 'expfunc', None):
        args.command.print_help()
    else:
        args.expfunc(args, rest)


def non_empty(value, question, keep=None):
    """Ask value while it is empty"""
    if value != "<ask>":
        if value:
            return value
        if keep is not None:
            return keep
    elif keep is None:
        return value
    while True:
        value = input("=> " + question + "\n")
        check_value = value.strip()
        if check_value:
            return value
        print("Please write a non-empty value")


def validate_yes_no(value):
    """Validate yes-no question"""
    if isinstance(value, bool):
        return value
    if value is None or value == "<ask>":
        return value
    check_value = value.strip().lower()
    if check_value in ("", "y", "yes", "t", "true", "1"):
        return True
    if check_value in ("n", "no", "f", "false", "0"):
        return False
    return None


def yes_no(value, question, keep=None):
    """Ask yes-no question while it is not valid"""
    if value != "<ask>":
        if value is not None:
            return value
        if keep is not None:
            return keep
    else:
        if keep is None:
            return True

    while True:
        value = input("=> " + question + " [Y/n] ")
        check_value = validate_yes_no(value)
        if check_value is not None:
            if value.strip() == "":
                print("Using default (yes)")
            return check_value
        print("Please respond with 'yes' or 'no'")

def start_experiment_cmd(args, _):
    """Start experiment"""
    data = load_home_config()
    experiment = data["experiment"]
    experiment["id"] = non_empty(
        args.id,
        "What is your participant ID or email address?",
        experiment["id"] if args.keep else None
    )
    experiment["lintingMessage"] = yes_no(
        args.linting_messages,
        "Can we collect linting messages that appear for you?",
        experiment["lintingMessage"] if args.keep else None
    )
    if experiment["lintingMessage"]:
        experiment["lintingTypes"] = True
    else:
        experiment["lintingTypes"] = yes_no(
            args.linting_types,
            "Can we collect the types of linting messages?",
            experiment["lintingTypes"] if args.keep else None
        )
    experiment["name"] = yes_no(
        args.name,
        "Can we collect the notebook name?",
        experiment["name"] if args.keep else None
    )
    experiment["code"] = yes_no(
        args.code,
        "Can we collect the notebook code?",
        experiment["code"] if args.keep else None
    )
    experiment["execution"] = yes_no(
        args.execution,
        "Can we collect execution information "
        "(i.e., size of executed cells, type of execution output)?",
        experiment["execution"] if args.keep else None
    )
    experiment["activity"] = yes_no(
        args.activity,
        "Can we collect activity information "
        "(i.e., Julynter filters, notebook opening and closing)?",
        experiment["activity"] if args.keep else None
    )
    experiment["sendServer"] = yes_no(
        args.send,
        "Can we send the collected data to a server on the fly?",
        experiment["sendServer"] if args.keep else None
    )
    experiment["server"] = (
        args.server
        or (experiment["server"] if args.keep else DEFAULT_EXPERIMENT_SERVER)
    )
    experiment["enabled"] = True
    save_home_config(data)
    data["header"] = "Configure"
    data["operation"] = "start"
    data["date"] = datetime.now().isoformat()
    log(data, "experiment", data)
    print("Experiment started! Do not forget to use `jupyter lab`"
          " or `julynter lab` during the next week")


def end_experiment_cmd(args, rest):
    """End experiment"""
    data = load_home_config()
    data['experiment']['enabled'] = False
    save_home_config(data)
    data['experiment']['enabled'] = True
    data["header"] = "Configure"
    data["operation"] = "stop"
    data["date"] = datetime.now().isoformat()
    log(data, "experiment", data)
    print("Experiment stoped!")

    if not args.prevent_zip:
        zip_experiment_cmd(args, rest)


def zip_experiment_cmd(args, _):
    """Zip command"""
    save_cwd = os.getcwd()
    zip_filename = os.path.abspath(args.file)
    archive_dir = os.path.dirname(zip_filename)
    root_dir = home_config_path()
    os.chdir(root_dir)
    base_dir = os.curdir
    try:
        if archive_dir and not os.path.exists(archive_dir):
            os.makedirs(archive_dir)
        with zipfile.ZipFile(zip_filename, "w",
                             compression=zipfile.ZIP_DEFLATED) as zfil:
            path = os.path.normpath(base_dir)
            if path != os.curdir:
                zfil.write(path, path)
            for dirpath, dirnames, filenames in os.walk(base_dir):
                for name in sorted(dirnames):
                    path = os.path.normpath(os.path.join(dirpath, name))
                    if not args.ignore_sent or not path.startswith('sent_'):
                        zfil.write(path, path)
                for name in filenames:
                    path = os.path.normpath(os.path.join(dirpath, name))
                    if not args.ignore_sent or not path.startswith('sent_'):
                        if os.path.isfile(path):
                            zfil.write(path, path)
    finally:
        os.chdir(save_cwd)

    print("Experiment pack {} created!".format(args.file))
    print("Please, send it to https://forms.gle/HmcP47AJoKMmmUms7")


def create_subparsers(subparsers):
    """Create experiment subparsers"""
    expparser = subparsers.add_parser(
        'experiment', help="Configure Julynter experiment"
    )
    expparser.set_defaults(func=base_experiment_cmd, command=expparser)
    expparser_sub = expparser.add_subparsers()

    expparser_start = expparser_sub.add_parser(
        "start", help="Start Julynter experiment"
    )
    expparser_start.set_defaults(expfunc=start_experiment_cmd)
    expparser_start.add_argument(
        '-k', '--keep', default=None, action="store_true",
        help='Keep old values'
    )
    expparser_start.add_argument(
        '-i', '--id', default=None, nargs='?', const="<ask>",
        help='Experiment id'
    )
    expparser_start.add_argument(
        '-m', '--linting-messages', type=validate_yes_no,
        nargs='?', default=None, const="<ask>",
        help='Collect linting messages'
    )
    expparser_start.add_argument(
        '-t', '--linting-types', type=validate_yes_no, nargs='?',
        default=None, const="<ask>",
        help='Collect linting types'
    )
    expparser_start.add_argument(
        '-a', '--activity', type=validate_yes_no, nargs='?',
        default=None, const="<ask>",
        help='Collect activity'
    )
    expparser_start.add_argument(
        '-e', '--execution', type=validate_yes_no, nargs='?',
        default=None, const="<ask>",
        help='Collect execution'
    )
    expparser_start.add_argument(
        '-n', '--name', type=validate_yes_no, nargs='?',
        default=None, const="<ask>",
        help='Collect notebook name'
    )
    expparser_start.add_argument(
        '-c', '--code', type=validate_yes_no, nargs='?',
        default=None, const="<ask>",
        help='Collect notebook code'
    )
    expparser_start.add_argument(
        '-s', '--send', type=validate_yes_no, nargs='?',
        default=None, const="<ask>",
        help='Send experiment to server'
    )
    expparser_start.add_argument(
        '--server', default=None,
        help='Experiment server'
    )


    expparser_stop = expparser_sub.add_parser(
        "stop", help="Stop Julynter experiment"
    )
    expparser_stop.set_defaults(expfunc=end_experiment_cmd)
    expparser_stop.add_argument(
        '-z', '--prevent-zip', action="store_true",
        help='Prevent it from generating pack'
    )
    expparser_stop.add_argument(
        '-f', '--file', type=str, default="julynter_experiment.zip",
        help='Output path for experiment pack'
    )
    expparser_stop.add_argument(
        '-s', '--ignore-sent', action="store_true",
        help='Ignore sent files'
    )


    expparser_zip = expparser_sub.add_parser("zip", help="Zip current experiment results")
    expparser_zip.set_defaults(expfunc=zip_experiment_cmd)
    expparser_zip.add_argument(
        '-f', '--file', type=str, default="julynter_experiment.zip",
        help='Output path for experiment pack'
    )
    expparser_zip.add_argument(
        '-s', '--ignore-sent', action="store_true",
        help='Ignore sent files'
    )



"""Define main runner class"""
import sys
import traceback
import time
import re

from copy import deepcopy
from math import ceil

import nbformat
import nbconvert
from jupyter_client.kernelspec import find_kernel_specs

from ..util import vprint, to_unicode, TimeoutException, Path
from .preprocessors import create_preprocessor
from .compare import cell_diff, DEFAULT_NORMALIZATION, DEFAULT_SIMILARITY


class StopRunException(Exception):
    """Represents an interruption of the execution"""

def clean_fail():
    """Return clean fail result dict"""
    return {
        "msg": None,
        "reason": None,
    }

def clean_result():
    """Return clean result dict"""
    return {
        "cell_order": [],
        "executed_cells": 0,
        "status": "not-run", # not-run, skipped, error, run
        "processed": ["attempt"], # attempt, loaded, timeout, exception
        "timeout": None,
        "duration": None,
        "last_cell_index": None,
        "count": None,
    }

def clean_diff_result():
    """Return clean diff result dict"""
    return {
        "diff": None,
        "diff_count": None,
        "diffnorm": None,
        "diffnorm_count": None,
        "processed": [], # finished, same-results, mismatch-results, same-norm, mismatch-norm
        "similarities": [],
    }


class Runner(object):
    """Run Jupyter notebooks"""
    # pylint: disable=useless-object-inheritance, too-many-instance-attributes

    def __init__(
            self, path, order, unsafe,
            kernel=None, force_fail=False, notebook_timeout=300, show_report=False,
            normalizations=DEFAULT_NORMALIZATION, calculate_similarity=DEFAULT_SIMILARITY,
            vindex=3
    ):
        # pylint: disable=dangerous-default-value, too-many-arguments
        self.kernel = kernel
        self.force_fail = force_fail
        self.notebook_timeout = notebook_timeout
        self.show_report = show_report
        self.normalizations = normalizations
        self.vindex = vindex
        self.calculate_similarity = calculate_similarity
        self.preprocessor = create_preprocessor(order, unsafe, vindex)
        self.path = Path(path).expanduser()
        self.notebook = None
        self.old_nb = None
        self.start_time = None

        self.result = clean_result()
        self.fail = clean_fail()
        self.diff_result = clean_diff_result()

    def _timeout_func(self, cell):
        """Define cell timeout"""
        # pylint: disable=unused-argument
        available = self.notebook_timeout - int(ceil(time.time() - self.start_time))
        return max(1, available)

    def load_file(self):
        """Load .ipynb file"""
        vprint(self.vindex, u"Reading file {}".format(to_unicode(self.path)))
        try:
            with open(str(self.path)) as fil:
                self.notebook = nbformat.read(fil, as_version=4)
            self.old_nb = deepcopy(self.notebook)
            self.update_processed("loaded")
        except OSError:
            vprint(self.vindex + 1, "Failed to open file")
            self.report_exit(
                "<Read notebook error>",
                traceback.format_exc()
            )

    def find_kernel(self):
        """Decide which kernel to use"""
        kernel_specs = find_kernel_specs()
        kernel = self.kernel
        if kernel:
            if kernel in kernel_specs:
                return kernel
            if self.force_fail:
                self.report_exit(
                    "<No available kernel>",
                    "Kernel {} not found".format(kernel)
                )
            else:
                vprint(self.vindex, (
                    "Configured Kernel {} not found. "
                    "Trying to detect other kernel"
                ).format(self.kernel))

        kernel = (
            self.notebook
            .get('metadata', {})
            .get('kernelspec', {})
            .get('name', 'python')
        )
        if kernel in kernel_specs:
            return kernel
        kernel = 'python'
        if kernel in kernel_specs:
            return kernel
        kernel = 'python{}'.format(sys.version_info[0])
        if kernel in kernel_specs:
            return kernel
        if kernel_specs:
            return next(iter(kernel_specs))
        return None

    def set_kernel(self, preprocessor):
        """Set preprocessor kernel"""
        kernel = self.find_kernel()
        if kernel is None:
            vprint(self.vindex, "No kernel found")
            self.report_exit(
                "<No available kernel>",
                "Kernel not found"
            )
        vprint(self.vindex, "Using kernel {}".format(kernel))
        preprocessor.kernel_name = kernel

    def execute_notebook(self, preprocessor):
        """Execute notebook"""
        preprocessor.timeout_func = self._timeout_func
        self.start_time = time.time()
        preprocessor.log.propagate = False
        preprocessor.preprocess(self.notebook, {'metadata': {'path': str(self.path.parent)}})
        preprocessor.log.propagate = True

    def run(self, clean=True):
        """Run notebook"""
        if clean:
            self.result = clean_result()
            self.fail = clean_fail()
        try:
            # pylint: disable=duplicate-except
            self.load_file()
            preprocessor = self.preprocessor()
            preprocessor.last_try = (-1, -1)
            self.set_kernel(preprocessor)
            skip = preprocessor.prepare_notebook_order(self.notebook, self.vindex)
            self.result["cell_order"] = preprocessor.cell_order
            if skip != "":
                vprint(self.vindex, "Skipping notebook. Reason: {}".format(skip))
                self.report_exit("<Skipping notebook>", skip, statuscode="skipped")

            timeout = 0
            try:
                vprint(self.vindex, "Executing notebook")
                self.execute_notebook(preprocessor)
                vprint(self.vindex + 1, "Finished")
            except TimeoutException:
                timeout = 1
                vprint(self.vindex + 1, "Timeout")
                self.update_processed("timeout")
            except RuntimeError:
                reason = "RuntimeError"
                vprint(self.vindex + 1, "Exception: {}".format(reason))
                self.update_processed("exception")
                self.update_reason(reason, traceback.format_exc())
            except AttributeError:
                reason = "Malformed Notebook"
                vprint(self.vindex + 1, "Exception: {}".format(reason))
                self.update_processed("exception")
                self.update_reason(reason, traceback.format_exc())
            except nbconvert.preprocessors.execute.CellExecutionError as exc:
                try:
                    reason = re.findall(r"\n(.*): .*\n$", str(exc))[-1]
                except IndexError:
                    reason = "<Unknown exception>"
                vprint(self.vindex + 1, "Exception: {}".format(reason))
                self.update_processed("exception")
                self.update_reason(reason, traceback.format_exc())

            vprint(self.vindex + 1, "Run up to {}".format(preprocessor.last_try))
            self.update_results(
                timeout=self.notebook_timeout,
                duration=time.time() - self.start_time,
                last_cell_index=preprocessor.last_try[1],
                count=preprocessor.last_try[0] + 1,
                executed_cells=preprocessor.last_try[0] + 1 - timeout,
                status="run"
            )
        except StopRunException:
            return False
        return True

    def compare(self, clean=True):
        """Compare notebook results"""
        if clean:
            self.diff_result = clean_diff_result()
        vprint(self.vindex, "Comparing notebooks")
        diff = []
        new_diff = []
        for _, index in zip(range(self.result["executed_cells"]), self.result["cell_order"]):
            vprint(self.vindex + 1, "Comparing cell {}".format(index))
            old_cell = self.old_nb.cells[index]
            new_cell = self.notebook.cells[index]
            original_equal, any_equal, diff_result = cell_diff(
                index, old_cell, new_cell, self.show_report,
                self.normalizations, self.calculate_similarity,
                self.vindex + 2
            )
            if not original_equal:
                diff.append(index)
            if not any_equal:
                new_diff.append(index)
            self.add_similarity(dict(
                index=index,
                **diff_result
            ))

        if not diff:
            vprint(self.vindex + 1, "Identical results")
            self.update_processed("same-results", usediff=True)
            self.update_results(
                usediff=True,
                diff="",
                diff_count=0
            )
        else:
            vprint(self.vindex + 1, "Diff on cells: {}".format(diff))
            self.update_processed("mismatch-results", usediff=True)
            self.update_results(
                usediff=True,
                diff=",".join(map(str, diff)),
                diff_count=len(diff)
            )

        if not new_diff:
            vprint(self.vindex + 1, "Identical results after normalizations")
            self.update_processed("same-norm", usediff=True)
            self.update_results(
                usediff=True,
                diffnorm="",
                diffnorm_count=0
            )
        else:
            vprint(self.vindex + 1, "Diff on cells after normalizations: {}".format(new_diff))
            self.update_processed("mismatch-norm", usediff=True)
            self.update_results(
                usediff=True,
                diffnorm=",".join(map(str, new_diff)),
                diffnorm_count=len(new_diff)
            )
        self.update_processed("finished", usediff=True)

    def save(self, output):
        """Save output notebook"""
        with open(str(output), "w") as fil:
            nbformat.write(self.notebook, fil)

    def update_processed(self, flag, usediff=False):
        """Add flag to processed"""
        res = self.diff_result if usediff else self.result
        res["processed"].append(flag)

    def update_reason(self, reason, msg):
        """Update execution reason and msg"""
        self.fail["reason"] = reason
        self.fail["msg"] = msg

    def update_results(self, usediff=False, **kwargs):
        """Update execution results"""
        res = self.diff_result if usediff else self.result
        for key, value in kwargs.items():
            if key in res:
                res[key] = value
            else:
                self.report_exit("<runner bug>", "Key {} not found in result dict".format(key))

    def add_similarity(self, sim):
        """Add similarity result to execution result"""
        self.diff_result["similarities"].append(sim)

    def report_exit(self, reason, msg, statuscode="error"):
        """Stop execution and give a reason"""
        self.update_reason(reason, msg)
        self.result["status"] = statuscode
        raise StopRunException()
